# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import fp_numerics
import jax
import jax.numpy as jnp

jax.config.update("jax_numpy_rank_promotion", "raise")


def assert_equal(x, expected_x, error_msg="", aux_x=None, mask=None):
  x = jnp.array(x)
  expected_x = jnp.array(expected_x)
  sh = x.shape
  assert expected_x.shape == sh, f"{sh=}, {expected_x.shape=}"
  if mask is not None:
    assert mask.shape == sh, f"{sh=}, {mask.shape=}"
    x = jnp.where(mask, x, jnp.zeros_like(x))
    expected_x = jnp.where(mask, expected_x, jnp.zeros_like(expected_x))
  if aux_x is not None:
    assert aux_x.shape == sh, f"{sh=}, {aux_x.shape=}"
  if len(sh) > 1:
    assert len(sh) == 2
    x = x.reshape((sh[0] * sh[1],))
    expected_x = expected_x.reshape((sh[0] * sh[1],))
    if aux_x is not None:
      aux_x = aux_x.reshape((sh[0] * sh[1],))
  (all_count,) = x.shape

  correct_val = x == expected_x
  correct_nan = jnp.isnan(x) & jnp.isnan(expected_x)
  correct = correct_val | correct_nan
  correct_count = jnp.sum(correct)
  bad = ~correct
  index_bad = (jnp.arange(all_count))[bad]
  values_bad = x[bad]
  expected_x_bad = expected_x[bad]
  error = (
      error_msg
      + "\n"
      + f"correct_count     = {correct_count}/{all_count}\n"
      f"!eq index         = {index_bad}\n"
      f"!eq lhs(actual)   = {values_bad}\n"
      f"!eq rhs(expected) = {expected_x_bad}\n"
  )
  if aux_x is not None:
    aux_x_bad = aux_x[bad]
    error += f"aux_x_bad          = {aux_x_bad}\n"
  assert correct_count == all_count, error


def fp_values(cfg: fp_numerics.FpNumericsConfig):
  nexp = cfg.nexp
  minexp = cfg.minexp
  nmant = cfg.nmant
  has_subnormals = cfg.has_subnormals
  has_two_nan = cfg.has_two_nan
  has_naninf = cfg.has_naninf

  assert not (has_two_nan and has_naninf)
  bits = 1 + nexp + nmant

  bitmasks = jnp.arange(2**bits, dtype=jnp.uint16)
  sign_bits = bitmasks >> (nexp + nmant)
  exp_bits = (bitmasks >> nmant) & ((1 << nexp) - 1)
  man_bits = bitmasks & ((1 << nmant) - 1)

  sign = sign_bits.astype(jnp.float32) * -2.0 + 1
  exp = exp_bits.astype(jnp.float32)
  man = man_bits.astype(jnp.float32)

  if cfg == fp_numerics.RADIX4:
    values = 4 ** jnp.array([-3, -2, -1, 0, 1, 2, 3], dtype=jnp.bfloat16)
    values = jnp.concatenate([jnp.zeros(1), values, -jnp.zeros(1), -values])
    return bitmasks, values

  if has_subnormals:
    exp -= 1  # resereve exp=0

  values = sign * 2 ** (exp + minexp) * (1 + man / 2**nmant)
  if has_subnormals:
    subnormals = sign * 2**minexp * (man / 2**nmant)
    values = jnp.where(exp_bits == 0, subnormals, values)
  if has_two_nan:
    is_nan = (exp_bits == 2**nexp - 1) & (man == 2**nmant - 1)
    values = jnp.where(is_nan, jnp.nan, values)
  if has_naninf:
    is_nan = exp_bits == 2**nexp - 1
    values = jnp.where(is_nan, jnp.nan, values)
    is_inf = (exp_bits == 2**nexp - 1) & (man == 0)
    values = jnp.where(is_inf, sign * jnp.inf, values)

  return bitmasks, values


class FpTest(parameterized.TestCase):

  @parameterized.parameters([
      dict(
          dtype=jnp.float8_e4m3fn,
          cfg=fp_numerics.float8_e4m3fn,
      ),
      dict(
          dtype=jnp.float8_e5m2,
          cfg=fp_numerics.float8_e5m2,
      ),
      dict(
          dtype=jnp.float16,
          cfg=fp_numerics.float16,
      ),
  ])
  def test_fp_values_on_extisting_dtypes(
      self,
      *,
      dtype,
      cfg,
  ):
    def fp_values_dtype(dtype):
      bits = jnp.finfo(dtype).bits
      match bits:
        case 8:
          mask_dtype = jnp.uint8
        case 16:
          mask_dtype = jnp.uint16
        case 32:
          mask_dtype = jnp.uint32
        case 64:
          mask_dtype = jnp.uint64
        case _:
          raise ValueError(f"Unsupported dtype: {dtype}")
      bitmasks = jnp.arange(2**bits, dtype=mask_dtype)
      values = jax.lax.bitcast_convert_type(bitmasks, dtype)
      values = values.astype(jnp.float32)
      return bitmasks, values

    nexp = jnp.finfo(dtype).nexp
    minexp = jnp.finfo(dtype).minexp
    nmant = jnp.finfo(dtype).nmant
    assert nexp == cfg.nexp
    assert minexp == cfg.minexp
    assert nmant == cfg.nmant
    bitmasks, values = fp_values(cfg)
    bitmasks1, expected_values = fp_values_dtype(dtype)
    assert (bitmasks == bitmasks1).all()
    assert_equal(
        values,
        expected_values,
        f" {cfg=}",
    )

  @parameterized.parameters([
      dict(
          nexp=3, minexp=0, nmant=0, has_subnormals=False,
          expected_values=[
              1, 2, 4, 8, 16, 32, 64, 128,
              -1, -2, -4, -8, -16, -32, -64, -128,
          ],
      ),
      dict(
          nexp=3, minexp=0, nmant=0, has_subnormals=True,
          expected_values=[
              0, 1, 2, 4, 8, 16, 32, 64,
              -0, -1, -2, -4, -8, -16, -32, -64,
          ],
      ),
      dict(
          nexp=3, minexp=-2, nmant=0, has_subnormals=False,
          expected_values=[
              0.25, 0.5, 1, 2, 4, 8, 16, 32,
              -0.25, -0.5, -1, -2, -4, -8, -16, -32,
          ],
      ),
      dict(
          nexp=3, minexp=-2, nmant=0, has_subnormals=True,
          expected_values=[
              0, 0.25, 0.5, 1, 2, 4, 8, 16,
              -0, -0.25, -0.5, -1, -2, -4, -8, -16,
          ],
      ),
      dict(
          nexp=2, minexp=0, nmant=0, has_subnormals=False,
          expected_values=[
              1, 2, 4, 8,
              -1, -2, -4, -8,
          ],
      ),
      dict(
          nexp=2, minexp=0, nmant=1, has_subnormals=False,
          expected_values=[
              1, 1.5, 2, 3, 4, 6, 8, 12,
              -1, -1.5, -2, -3, -4, -6, -8, -12,
          ],
      ),
      dict(
          nexp=2, minexp=0, nmant=1, has_subnormals=True,
          expected_values=[
              0, 0.5, 1, 1.5, 2, 3, 4, 6,
              -0, -0.5, -1, -1.5, -2, -3, -4, -6,
          ],
      ),
      dict(
          nexp=1, minexp=0, nmant=2, has_subnormals=True,
          expected_values=[
              0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75,
              -0, -0.25, -0.5, -0.75, -1, -1.25, -1.5, -1.75,
          ],
      ),
      dict(
          nexp=0, minexp=0, nmant=3, has_subnormals=True,
          expected_values=[
              0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875,
              -0, -0.125, -0.25, -0.375, -0.5, -0.625, -0.75, -0.875,
          ],
      ),
      dict(
          nexp=3, minexp=0, nmant=0, has_subnormals=True,
          expected_values=[
              0, 4**-3, 4**-2, 4**-1, 4**0, 4**1, 4**2, 4**3,
              -0, -4**-3, -4**-2, -4**-1, -4**0, -4**1, -4**2, -4**3,
          ],
          radix=4,
      ),
  ])  # pyformat: disable
  def test_fp_some_fp_values(
      self,
      nexp,
      minexp,
      nmant,
      has_subnormals,
      expected_values,
      radix=2,
  ):
    cfg = fp_numerics.FpNumericsConfig(
        nexp=nexp,
        minexp=minexp,
        nmant=nmant,
        has_subnormals=has_subnormals,
        has_two_nan=False,
        has_naninf=False,
        radix=radix,
    )

    bitmasks, values = fp_values(cfg)
    bits = 1 + nexp + nmant
    assert (bitmasks == jnp.arange(2**bits)).all()
    assert_equal(values, expected_values, f"{nexp=}, {minexp=}, {nmant=}")

  # For now we just test:
  #   has_two_nan=False,
  #   has_naninf=False,
  @parameterized.parameters([
      # FP4
      dict(nexp=3, minexp=0, nmant=0, has_subnormals=False),
      dict(nexp=3, minexp=0, nmant=0, has_subnormals=True),
      # dict(nexp=3, minexp=-2, nmant=0, has_subnormals=False),
      # dict(nexp=3, minexp=-2, nmant=0, has_subnormals=True),
      dict(nexp=2, minexp=0, nmant=1, has_subnormals=False),
      dict(nexp=2, minexp=0, nmant=1, has_subnormals=True),
      dict(nexp=1, minexp=0, nmant=2, has_subnormals=False),
      dict(nexp=1, minexp=0, nmant=2, has_subnormals=True),
      dict(nexp=0, minexp=0, nmant=3, has_subnormals=False),
      dict(nexp=0, minexp=0, nmant=3, has_subnormals=True),
      # FP6
      dict(nexp=5, minexp=0, nmant=0, has_subnormals=True),
      dict(nexp=5, minexp=0, nmant=0, has_subnormals=False),
      dict(nexp=4, minexp=0, nmant=1, has_subnormals=True),
      dict(nexp=4, minexp=0, nmant=1, has_subnormals=False),
      dict(nexp=3, minexp=0, nmant=2, has_subnormals=True),
      dict(nexp=3, minexp=0, nmant=2, has_subnormals=False),
      dict(nexp=2, minexp=0, nmant=3, has_subnormals=True),
      dict(nexp=2, minexp=0, nmant=3, has_subnormals=False),
      dict(nexp=1, minexp=0, nmant=4, has_subnormals=True),
      dict(nexp=1, minexp=0, nmant=4, has_subnormals=False),
      dict(nexp=0, minexp=0, nmant=5, has_subnormals=True),
      dict(nexp=0, minexp=0, nmant=5, has_subnormals=False),
      # FP5
      dict(nexp=4, minexp=0, nmant=0, has_subnormals=True),
      dict(nexp=4, minexp=0, nmant=0, has_subnormals=False),
      dict(nexp=3, minexp=0, nmant=1, has_subnormals=True),
      dict(nexp=3, minexp=0, nmant=1, has_subnormals=False),
      dict(nexp=2, minexp=0, nmant=2, has_subnormals=True),
      dict(nexp=2, minexp=0, nmant=2, has_subnormals=False),
      dict(nexp=1, minexp=0, nmant=3, has_subnormals=True),
      dict(nexp=1, minexp=0, nmant=3, has_subnormals=False),
      dict(nexp=0, minexp=0, nmant=4, has_subnormals=True),
      dict(nexp=0, minexp=0, nmant=4, has_subnormals=False),
      # Misc
      dict(nexp=2, minexp=0, nmant=0, has_subnormals=False),
      dict(nexp=3, minexp=0, nmant=0, has_subnormals=True, radix=4),
  ])
  def test_fp_round(
      self,
      nexp,
      minexp,
      nmant,
      has_subnormals,
      radix=2,
      det_x_count=2**7,  # needs to be power of 2, test_noise_axis exact eq
      x_count=128,  # TODO(lew): Test fails for 100, why?
      sr_sample_count=1000000,
      # TODO(lew): Enable these:
      # x_count=2**12,  # needs to be power of 2
      # sr_sample_count=1000,
  ):
    det_x_count += 1  # endpoint
    x_count += 1  # endpoint
    cfg = fp_numerics.FpNumericsConfig(
        nexp=nexp,
        minexp=minexp,
        nmant=nmant,
        has_subnormals=has_subnormals,
        has_two_nan=False,
        has_naninf=False,
        radix=radix,
    )
    _, bucket_centers = fp_values(cfg)
    assert bucket_centers.dtype == jnp.float32
    # bucket_centers = bucket_centers.astype(jnp.float32)
    bucket_centers = bucket_centers.sort()
    assert bucket_centers[-1] == fp_numerics.fp_largest_representable(cfg), (
        bucket_centers[-1],
        fp_numerics.fp_largest_representable(cfg),
    )
    bucket_boundaries = (bucket_centers[1:] + bucket_centers[:-1]) / 2
    assert (
        -bucket_boundaries[0] == bucket_boundaries[-1]
    ), f"Symmetric datatype {bucket_boundaries[0]} vs {bucket_boundaries[-1]}"
    last_bucket_half_size = bucket_centers[-1] - bucket_boundaries[-1]
    assert last_bucket_half_size > 0
    edge_of_last_bucket = jnp.array(
        [bucket_centers[-1] + last_bucket_half_size]
    )
    bucket_boundaries = jnp.concatenate(
        [-edge_of_last_bucket, bucket_boundaries, edge_of_last_bucket]
    )

    # TEST: Check that each interval is rounded to the center.

    bucket_lower = bucket_boundaries[:-1]
    bucket_upper = bucket_boundaries[1:]
    if radix == 4:
      # For 4-bit radix4, the edge between bucket 4**-3 and bucket 0 is not
      # (4**-3 + 0) / 2. The edge is (4**-3 + 4**-4) / 2.
      # That is where floor(log4(1.6 * x)) isexactly -3.
      bucket_lower = jnp.where(
          jnp.abs(bucket_lower) == (4**-3 / 2),
          jnp.sign(bucket_lower) * (4**-3 + 4**-4) / 2,
          bucket_lower,
      )
      bucket_upper = jnp.where(
          jnp.abs(bucket_upper) == (4**-3 / 2),
          jnp.sign(bucket_upper) * (4**-3 + 4**-4) / 2,
          bucket_upper,
      )
    data = jnp.linspace(
        bucket_lower,
        bucket_upper,
        x_count,
        endpoint=True,
        axis=0,
        dtype=jnp.float32,
    )
    assert data.shape == (x_count, 2 ** (1 + nexp + nmant))
    # TODO(lew): Remove when fp_round supports float32. Also below.
    data = data.astype(jnp.bfloat16)
    bucked_interior_mask = jnp.logical_and(
        data != bucket_lower[jnp.newaxis, :],
        data != bucket_upper[jnp.newaxis, :],
    )
    qdata = fp_numerics.fp_round(
        data,
        cfg=cfg,
        key=None,
        stochastic_rounding=False,
    )
    assert qdata.shape == data.shape, f"{qdata.shape=}, {data.shape=}"
    bucket_centers_1 = jnp.expand_dims(bucket_centers, axis=0)
    bucket_centers_1 = jnp.broadcast_to(bucket_centers_1, qdata.shape)
    assert_equal(
        qdata,
        bucket_centers_1,
        f"Config: {nexp=}, {minexp=}, {nmant=}, {has_subnormals=}\n"
        + f"{bucket_centers=}\n"
        + f"{data=}\n"
        + f"{qdata=}\n\n",
        aux_x=data,
        mask=bucked_interior_mask,
    )

    # TEST: Check that overflows are clamped to edge_of_last_bucket.
    for sign in [-1, 1]:
      data = jnp.linspace(
          sign * edge_of_last_bucket,
          sign * 10 * edge_of_last_bucket,
          x_count,
          endpoint=True,
          dtype=jnp.float32,
      )
      data = data.astype(jnp.bfloat16)
      qdata = fp_numerics.fp_round(
          data,
          cfg=cfg,
          key=None,
          stochastic_rounding=False,
      )
      last_center = sign * bucket_centers[-1]

      assert_equal(
          qdata,
          jnp.full_like(qdata, last_center),
          f"Config: {nexp=}, {minexp=}, {nmant=}, {has_subnormals=}\n"
          + f"{bucket_centers=}\n"
          + f"{data=}\n"
          + f"{qdata=}\n\n",
          aux_x=data,
      )

    # TEST: Equi-spaced noise deterministic "SR"

    x = jnp.linspace(
        bucket_centers[0],
        bucket_centers[-1],
        det_x_count,
        endpoint=True,
        dtype=jnp.float32,
    ).astype(jnp.bfloat16)
    # The number of broadcasts should be consistent with the effective noise bit
    # For example, if using 16-bit noise, inputs should be replicated by 2**16.
    bx = jnp.broadcast_to(x[:, jnp.newaxis], (det_x_count, 2**16))
    deterministic_qx = fp_numerics.fp_round(
        bx.astype(jnp.bfloat16),
        cfg=cfg,
        key=None,
        test_noise_axis=1,
        stochastic_rounding=True,
    )
    mean_det_qx = jnp.mean(deterministic_qx, axis=1)
    assert_equal(mean_det_qx, x, f"{cfg=}")

    # TEST: Unbiased stochastic rounding

    x = jnp.linspace(
        bucket_centers[0],
        bucket_centers[-1],
        x_count,
        endpoint=True,
        dtype=jnp.float32,
    ).astype(jnp.bfloat16)
    bx = jnp.broadcast_to(x[:, jnp.newaxis], (x_count, sr_sample_count))
    qx = fp_numerics.fp_round(
        bx.astype(jnp.bfloat16),
        cfg=cfg,
        key=jax.random.PRNGKey(42),
        stochastic_rounding=True,
    )

    assert qx.dtype == jnp.bfloat16
    qx = qx.astype(jnp.float32)
    # qx can hit at most 2 values, larger and smaller bucket center
    qx_smaller = jnp.min(qx, axis=1, keepdims=True)
    qx_larger = jnp.max(qx, axis=1, keepdims=True)
    assert ((qx == qx_larger) | (qx == qx_smaller)).all()

    # Compute empirical probability of rounding up and down.
    qx_smaller_count = jnp.sum(qx == qx_smaller, axis=1, keepdims=True)
    qx_larger_count = jnp.sum(qx == qx_larger, axis=1, keepdims=True)
    total_count = qx_smaller_count + qx_larger_count
    assert_equal(qx_smaller != qx_larger, total_count == sr_sample_count, "zz")

    total_count = jnp.where(  # if we hit bucket center exactly.
        qx_smaller == qx_larger, total_count / 2, total_count
    )
    assert_equal(
        total_count,
        jnp.full_like(total_count, sr_sample_count),
        "",
    )
    p_smaller = qx_smaller_count / sr_sample_count
    bucket_size = qx_larger - qx_smaller
    expected_p_smaller = (qx_larger - x[:, jnp.newaxis]) / bucket_size
    expected_p_smaller = jnp.where(bucket_size == 0, 1.0, expected_p_smaller)
    p_err = jnp.abs(p_smaller - expected_p_smaller)
    p_stderr = p_err * jnp.sqrt(sr_sample_count)
    # print(f"{p_stderr=}")
    assert (p_stderr < 2.0).all(), p_stderr

  @parameterized.parameters([
      # FP4
      dict(cfg=fp_numerics.e3m0, sr=False),
      dict(cfg=fp_numerics.e2m1, sr=False),
      dict(cfg=fp_numerics.e1m2, sr=False),
      dict(cfg=fp_numerics.e0m3, sr=False),
      dict(cfg=fp_numerics.e3m0_ocp, sr=False),
      dict(cfg=fp_numerics.e2m1_ocp, sr=False),
      dict(cfg=fp_numerics.e1m2_ocp, sr=False),
      dict(cfg=fp_numerics.e0m3_ocp, sr=False),
      dict(cfg=fp_numerics.e3m0, sr=True),
      dict(cfg=fp_numerics.e2m1, sr=True),
      dict(cfg=fp_numerics.e1m2, sr=True),
      dict(cfg=fp_numerics.e0m3, sr=True),
      dict(cfg=fp_numerics.e3m0_ocp, sr=True),
      dict(cfg=fp_numerics.e2m1_ocp, sr=True),
      dict(cfg=fp_numerics.e1m2_ocp, sr=True),
      dict(cfg=fp_numerics.e0m3_ocp, sr=True),
      # FP5
      dict(cfg=fp_numerics.e4m0_ocp, sr=False),
      dict(cfg=fp_numerics.e3m1_ocp, sr=False),
      dict(cfg=fp_numerics.e2m2_ocp, sr=False),
      dict(cfg=fp_numerics.e1m3_ocp, sr=False),
      dict(cfg=fp_numerics.e0m4_ocp, sr=False),
      dict(cfg=fp_numerics.e4m0_ocp, sr=True),
      dict(cfg=fp_numerics.e3m1_ocp, sr=True),
      dict(cfg=fp_numerics.e2m2_ocp, sr=True),
      dict(cfg=fp_numerics.e1m3_ocp, sr=True),
      dict(cfg=fp_numerics.e0m4_ocp, sr=True),
      # FP6
      dict(cfg=fp_numerics.e5m0_ocp, sr=False),
      dict(cfg=fp_numerics.e4m1_ocp, sr=False),
      dict(cfg=fp_numerics.e3m2_ocp, sr=False),
      dict(cfg=fp_numerics.e2m3_ocp, sr=False),
      dict(cfg=fp_numerics.e1m4_ocp, sr=False),
      dict(cfg=fp_numerics.e0m5_ocp, sr=False),
      dict(cfg=fp_numerics.e5m0_ocp, sr=True),
      dict(cfg=fp_numerics.e4m1_ocp, sr=True),
      dict(cfg=fp_numerics.e3m2_ocp, sr=True),
      dict(cfg=fp_numerics.e2m3_ocp, sr=True),
      dict(cfg=fp_numerics.e1m4_ocp, sr=True),
      dict(cfg=fp_numerics.e0m5_ocp, sr=True),
      # minexp not implemented, hence the following cfgs are skipped
      # dict(cfg=fp_numerics.float8_e4m3fn),
      # dict(cfg=fp_numerics.float8_e5m2),
      # dict(cfg=fp_numerics.float16),
      # Misc
      dict(cfg=fp_numerics.e1m0, sr=False),
      dict(cfg=fp_numerics.e1m0, sr=True),
      dict(cfg=fp_numerics.e1m2, sr=False),
      dict(cfg=fp_numerics.e1m2, sr=True),
  ])
  def test_fp_round_old_vs_new(
      self, cfg: fp_numerics.FpNumericsConfig, sr: bool
  ):
    print(cfg)
    print(f"stochastic_rounding: {sr}")
    x = jnp.arange(0, 2**16, step=1, dtype=jnp.uint16)
    x = jax.lax.bitcast_convert_type(x, jnp.bfloat16)
    nan_mask = jnp.isnan(x)
    if sr:
      for key_num in [0, 12, 42, 17, 96, 101]:
        key = jax.random.PRNGKey(key_num)
        out = fp_numerics.fp_round(x, cfg=cfg, key=key, stochastic_rounding=sr)
        out_new = fp_numerics.fp_round_new(
            x, cfg=cfg, key=key, stochastic_rounding=sr
        )
        assert (out[~nan_mask] == out_new[~nan_mask]).all()
        assert jnp.isnan(out_new[nan_mask]).all()
    else:
      out = fp_numerics.fp_round(x, cfg=cfg, key=None, stochastic_rounding=sr)
      out_new = fp_numerics.fp_round_new(
          x, cfg=cfg, key=None, stochastic_rounding=sr
      )
      assert (out[~nan_mask] == out_new[~nan_mask]).all()
      assert jnp.isnan(out_new[nan_mask]).all()

  def test_e1m2_vs_e0m3(self):
    e1m2 = fp_numerics.FpNumerics(cfg=fp_numerics.e1m2_ocp)
    e0m3 = fp_numerics.FpNumerics(cfg=fp_numerics.e0m3_ocp)
    context = utils.Context(key=None, train_step=None)
    key = jax.random.PRNGKey(42)
    x = jax.random.normal(key, shape=(64, 64), dtype=jnp.bfloat16)
    x = x / jnp.abs(x).max()
    e1m2_input = x * e1m2.get_quant_bound()
    e0m3_input = x * e0m3.get_quant_bound()
    assert (e1m2_input == (e0m3_input * 2)).all()
    y_e1m2, _ = e1m2.vjp_fwd(e1m2_input, context=context)
    y_e0m3, _ = e0m3.vjp_fwd(e0m3_input, context=context)
    assert (y_e1m2 == y_e0m3 * 2).all()

if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to get bounds of activation/input using its statistics."""

import dataclasses
import typing
from typing import Iterable, Optional, Tuple, Union

from aqt.jax_legacy.jax import primitives
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax import shape_utils
from aqt.jax_legacy.jax.flax import struct as flax_struct
from aqt.jax_legacy.jax.stats import Stats
from flax import linen as nn
from jax import lax
import jax.numpy as jnp

dataclass = flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass


class DynamicBounds(nn.Module):
  """Get Bounds of activation using statistics.

  Attributes:
    hyper: hyperparamater to compute bound from statistics.
  """

  @dataclass
  class Hyper:
    """Hyperparameters for DynamicBounds."""
    granularity: quant_config.QuantGranularity
    # Clipping coefficient applied during dynamic quantization.
    # E.g. if it is 0.9 then dynamic scale will be reduced by 10%.
    clipping_coeff: float = 1.0

  @dataclass
  class Params:
    """Parameters for act quantiztaion using get_bounds."""
    # Axis along which to quantize activations.
    quant_axis: Optional[Iterable[int]] = None
    # Optional shape to verify if bounds shape is expected. Defaults to None.
    expected_bounds_shape: Union[None, int, Tuple[int, ...]] = None
    # Optional name of the get_bounds module.
    module_name: Optional[str] = None

  hyper: Hyper

  @nn.compact
  def __call__(
      self,
      x: jnp.ndarray,
      *,
      bounds_params: Params,
  ) -> jnp.ndarray:
    """Compute the input batch statistics.

    Args:
      x: the input to get bounds from using statistics.
      bounds_params: parameters to compute input's statistics and bounds.

    Returns:
      Bound value (same shape as inputs).
    """

    x = jnp.asarray(x, jnp.float32)

    hyper = self.hyper

    if bounds_params.quant_axis is not None:
      quant_axis = bounds_params.quant_axis
    else:
      if hyper.granularity == quant_config.QuantGranularity.PER_TENSOR:
        # Equivalently, this could be written as
        # quant_axis = tuple(range(x.ndim))
        quant_axis = None
      elif hyper.granularity == quant_config.QuantGranularity.PER_CHANNEL:
        # Quantize by aggregating activation statistics across all dimensions of
        # the activation tensor EXCEPT the batch dimension. Also known as per
        # example statistics
        quant_axis = tuple(range(1, x.ndim))
      else:
        raise ValueError(f'Unknown granularity {hyper.granularity}')

    abs_max_x = primitives.max_abs_weights(x, axis=quant_axis)

    return abs_max_x * self.hyper.clipping_coeff


class GetBounds(nn.Module):
  """Get Bounds of activation using statistics.

  Attributes:
    hyper: hyperparamater to compute bound from statistics.
  """

  @dataclass
  class Hyper:
    """Hyperparameters for GetBounds."""
    # bound =
    # stddev_coeff * mix_coeff * stddev + absdev_coeff * (1-mix_coeff) * absdev
    initial_bound: float  # Initial bounds value before bounds get updated for
    # the first time
    granularity: quant_config.QuantGranularity
    stddev_coeff: float  # param to weigh the stddev
    absdev_coeff: float  # param to weigh the absdev, not used in ucb
    mix_coeff: float  # 0 = only absdev, 1 = only stddev, not used in ucb
    reset_stats: bool = False  # whether to reset stats when bounds are updated
    # Exponential moving average coefficient for stats collection. 'None'
    # indicates to use a simple mean of a set of past samples, as governed by
    # the 'reset_stats' parameter.
    ema_coeff: Optional[float] = None

    # TODO(shivaniagrawal): Refactor these boolean flags to an enum since
    # they are mututally exclusive.

    # Whether to use |mean| + centered stddev formula to calculate bounds.
    # CAMS is an acronym for Centered Absolute Mean + Stddev.
    use_cams: bool = False
    # Whether to exclude zeros from statistics computation. This can be useful
    # when zeros were likely introduced by preceding ops, e.g. relu or dropout.
    exclude_zeros: bool = False
    # Whether to use a running mean of the maximum of the absolute value of
    # activation tensors to calculate bounds.
    use_mean_of_max: bool = False

    # below four variables are new get_bound coefficients
    # coefficients should be float, but set to None as default value
    # to avoid invisible bugs
    cams_coeff: Optional[float] = None
    cams_stddev_coeff: Optional[float] = None
    mean_of_max_coeff: Optional[float] = None
    fixed_bound: Optional[float] = None
    # whether to use old or new code to compute the bound
    use_old_code: bool = True

  @dataclass
  class Params:
    """Parameters for act quantiztaion using get_bounds."""
    update_stats: bool  # Whether to update statistics.
    update_bounds: bool  # Whether to update bounds.
    # The axis name used to combine batch statistics from multiple devices.
    # See `jax.pmap` for a description of axis names. Defaults to None.
    paxis_name: Optional[str] = None
    # Optional shape to verify if bounds shape is expected. Defaults to None.
    expected_bounds_shape: Union[None, int, Tuple[int, ...]] = None
    # Optional boolean tensor of the same shape as 'x' specifying which values
    # of 'x' to use as part of the bounds calculation. 'True' indicates the
    # corresponding value from 'x' should be used. If None, all values are used.
    mask: Optional[jnp.ndarray] = None
    # Optional name of the get_bounds module.
    module_name: Optional[str] = None

  hyper: Hyper

  @nn.compact
  def __call__(
      self,
      x: jnp.ndarray,
      *,
      bounds_params: Params,
  ) -> jnp.ndarray:
    """Compute the input batch statistics.

    Args:
      x: the input to get bounds from using statistics.
      bounds_params: parameters to compute input's statistics and bounds.

    Returns:
      Bound value (same shape as inputs).
    """

    if bounds_params.mask is not None:
      shape_utils.assert_shapes_compatible(x.shape, bounds_params.mask.shape)

    x = jnp.asarray(x, jnp.float32)

    hyper = self.hyper

    is_initializing = not self.has_variable('get_bounds', 'stats')

    if hyper.granularity == quant_config.QuantGranularity.PER_TENSOR:
      # Equivalently, this could be written as
      # quant_axis = tuple(range(x.ndim))
      quant_axis = None
      stats_shape = (1,) * len(x.shape)
    elif hyper.granularity == quant_config.QuantGranularity.PER_CHANNEL:
      # Quantize by aggregating activation statistics across all dimensions of
      # the activation tensor EXCEPT the last dimension, which we interpret as
      # the channel dimension. For example, in a transformer context, x might
      # have a shape corresponding to [example, token, channel], in which case
      # this aggregates activation statistics separately for each feature, where
      # for each feature it aggregates over all unmasked tokens in all examples.
      quant_axis = tuple(range(x.ndim - 1))
      stats_shape = (1,) * (x.ndim - 1) + (x.shape[-1],)
    else:
      raise ValueError(f'Unknown granularity {hyper.granularity}')

    stats_state = self.variable('get_bounds', 'stats', Stats.stats_initializer,
                                stats_shape)

    def bound_initializer(shape):
      return hyper.initial_bound * jnp.ones(shape)

    bounds = self.variable('get_bounds', 'bounds', bound_initializer,
                           stats_shape)

    if bounds_params.update_stats and not is_initializing:
      stats_state.value = Stats.create_updated_stats(
          stats_state.value,
          x,
          mask=bounds_params.mask,
          axis=quant_axis,
          paxis_name=bounds_params.paxis_name,
          alpha=hyper.ema_coeff,
          exclude_zeros=hyper.exclude_zeros)

    if bounds_params.update_bounds and not is_initializing:
      bounds.value = self._stats_to_bounds(stats_state.value)
      if hyper.reset_stats:
        stats_state.value = Stats.stats_initializer(stats_shape)
    return bounds.value

  def _stats_to_bounds(self, stats_value: Stats) -> jnp.ndarray:
    """Computes activation clipping bounds from activation statistics."""
    hyper = self.hyper
    maximum = stats_value.mean_batch_maximum
    minimum = stats_value.mean_batch_minimum
    mom = jnp.maximum(jnp.abs(maximum), jnp.abs(minimum))
    stddev_uncentered = lax.sqrt(stats_value.mean_sq)
    absdev_uncentered = stats_value.mean_abs
    stddev = lax.sqrt(stats_value.mean_sq - stats_value.mean**2)
    abs_mean = jnp.abs(stats_value.mean)
    if hyper.use_old_code:  # old code of computing the bound
      if hyper.use_cams:  # upper confidence bound formula
        return abs_mean + hyper.stddev_coeff * stddev
      elif hyper.use_mean_of_max:
        return mom
      else:
        return (hyper.mix_coeff * hyper.stddev_coeff * stddev_uncentered +
                (1 - hyper.mix_coeff) * hyper.absdev_coeff * absdev_uncentered)
    else:  # use new way of computing the bound
      cams = abs_mean + hyper.cams_stddev_coeff * stddev
      return (hyper.fixed_bound + hyper.mean_of_max_coeff * mom +
              hyper.stddev_coeff * stddev_uncentered +
              hyper.absdev_coeff * absdev_uncentered + hyper.cams_coeff * cams)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""dot_general implementation for Pallas."""

from aqt.jax.v2 import aqt_dot_general
from aqt.jax.v2 import aqt_tensor
import jax
import jax.numpy as jnp

QTensor = aqt_tensor.QTensor
DequantMode = aqt_dot_general.DequantMode


# TODO(wppark): Match signature of load_qtensor with pl.load to support
# partial loading of QTensor.
def load_qtensor(qt: QTensor) -> QTensor:
  """Materialize QTensor of MemoryRef of pallas into QTensor of jax.Array."""

  if qt.qvalue is not None:
    qt.qvalue = qt.qvalue[...]
  if qt.scale is not None:
    qt.scale = [s[...] for s in qt.scale]
  if qt.scale_t is not None:
    qt.scale_t = [st[...] for st in qt.scale_t]

  return qt


def _dtype_to_bits(dtype) -> None | int:
  if dtype in [jnp.bfloat16, jnp.float32]:
    return None
  if dtype == jnp.int8:
    return 8
  elif dtype == jnp.int4:
    return 4
  else:
    raise ValueError(
        'dtype must be one one of {jnp.bfloat16, jnp.float32, jnp.int8,'
        ' jnp.int4}'
    )


def dot_general(
    lhs: QTensor | jax.Array,
    rhs: QTensor | jax.Array,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    precision=None,
    preferred_element_type=None,
    lhs_dequant_mode: DequantMode = DequantMode.OUTPUT,
    rhs_dequant_mode: DequantMode = DequantMode.OUTPUT,
) -> jax.Array:
  """lax.dot_general replacement for pallas which always returns dequantized output.

  Args:
    lhs: left hand side tensor.
    rhs: right hand side tensor.
    dimension_numbers: dimension numbers for dot_general.
    precision: this is ignored, but added to match the signature of
      jax.lax.dot_general.
    preferred_element_type: preferred output element type after dequantization.
    lhs_dequant_mode: This decides where lhs is dequantized. Default is OUTPUT
      where dequantization is applied to the output of dot_general. OTHER_INPUT
      applies dequantization to the other input before dot_general and
      THIS_INPUT applies dequantization to the current input before dot_general.
    rhs_dequant_mode: This decides where rhs is dequantized. Default is OUTPUT.

  Returns:
    Dequantized output of dot_general.
  """
  # precision is not used.
  del precision

  # The code below is supposed to be executed inside pallas kernel.

  # When only one of operands is quantized, Jax implicitly cast int8 into float
  # and performs dot_general. However, pallas requires explicit casting when
  # only one of operands is quantized.
  is_both_quantized = isinstance(lhs, QTensor) and isinstance(rhs, QTensor)
  if isinstance(lhs, QTensor) and not is_both_quantized:
    promoted_dtype = jnp.promote_types(lhs.dequant_dtype, rhs)
    lhs.qvalue = lhs.qvalue.astype(promoted_dtype)
  if isinstance(rhs, QTensor) and not is_both_quantized:
    promoted_dtype = jnp.promote_types(rhs.dequant_dtype, lhs)
    rhs.qvalue = rhs.qvalue.astype(promoted_dtype)

  if isinstance(lhs, jax.Array):
    lhs = QTensor(
        qvalue=lhs, scale=[], scale_t=None, bias=[], dequant_dtype=lhs.dtype
    )
  if isinstance(rhs, jax.Array):
    rhs = QTensor(
        qvalue=rhs, scale=[], scale_t=None, bias=[], dequant_dtype=rhs.dtype
    )

  if preferred_element_type is None:
    preferred_element_type = jnp.promote_types(
        lhs.dequant_dtype, rhs.dequant_dtype
    )

  lhs_bits = _dtype_to_bits(lhs.qvalue.dtype)
  rhs_bits = _dtype_to_bits(rhs.qvalue.dtype)
  cfg = aqt_dot_general.dot_general_raw_make(
      lhs_bits=lhs_bits, rhs_bits=rhs_bits
  )

  cfg.lhs.dequant_mode = lhs_dequant_mode
  cfg.rhs.dequant_mode = rhs_dequant_mode
  # Accumulator dtype does not need to be modified. Unless both inputs are
  # dequantized on output.
  if (
      cfg.lhs.dequant_mode != DequantMode.OUTPUT
      or cfg.rhs.dequant_mode != DequantMode.OUTPUT
  ):
    cfg.dg_accumulator_dtype = None

  out_qtensor = aqt_dot_general._qtensor_dot_general(  # pylint: disable=protected-access
      lhs,
      rhs,
      dimension_numbers,
      cfg,
      dequant_dtype=preferred_element_type,
  )
  return out_qtensor.dequant()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Freezer for writing & storing general Flax structure."""

import enum
from typing import Any, Callable

from flax.core import meta as nn_meta
import flax.linen as nn


class FreezerMode(enum.Enum):
  NONE = 1
  WRITE = 2
  READ = 3


_FREEZE_VAR_NAME = 'frozen'


class Freezer(nn.Module):
  """Flax module that can freeze its input.

  On default it is an identity function that saves the input in a variable.
  In 'mode=FreezerMode.READ' mode, ignores the input and returns the frozen
  value. It is usefult to implement 'constant folding' and put quantized weights
  and scales in the checkpoint for serving. Specifically:

  self.get() returns None when freeze_mode=NONE or WRITE, returns variable when
  freeze_mode=READ.
  self.set() does nothing when freeze_mode=NONE or READ, creates and stores
  input value freeze_mode=CONVERT.
  """

  collection: str
  mode: FreezerMode
  axis_metadata_wrapper: None | Callable[..., nn_meta.AxisMetadata] = None

  @nn.compact
  def _get_or_set(self, inputs: Any, is_set: bool) -> None | Any:
    def initializer():
      if self.axis_metadata_wrapper is not None:
        return self.axis_metadata_wrapper(inputs)
      return inputs

    if is_set:
      match self.mode:
        case FreezerMode.NONE:
          pass
        case FreezerMode.WRITE:
          s = self.variable(self.collection, _FREEZE_VAR_NAME, initializer)
          s.value = inputs
          return None
        case FreezerMode.READ:
          # Set in READ mode works as an initializer for checkpoint reading.
          # we don't want to change the variable during the serving
          _ = self.variable(self.collection, _FREEZE_VAR_NAME, initializer)
          return None
        case _:
          # Nothing matched.
          assert False, 'Unknown quant mode.'
      return None
    else:
      match self.mode:
        case FreezerMode.NONE:
          return None
        case FreezerMode.WRITE:
          return None
        case FreezerMode.READ:
          if not self.has_variable(self.collection, _FREEZE_VAR_NAME):
            return None

          msg = 'Initialization should not happen in Get mode, but in Set mode.'

          def initializer_poison():
            assert False, msg

          return self.variable(
              self.collection, _FREEZE_VAR_NAME, initializer_poison
          ).value
        case _:
          # Nothing matched.
          assert False, 'Unknown quant mode.'

  def get(self) -> None | Any:
    return self._get_or_set(None, is_set=False)

  def set(self, inputs: Any) -> None | Any:
    return self._get_or_set(inputs, is_set=True)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration dataclasses."""

from typing import Literal, Sequence

from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import calibration
from aqt.jax.v2 import tiled_dot_general
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import fp8_numerics
from aqt.jax.v2.numerics import no_numerics
from aqt.jax.v2.numerics import numerics
from aqt.jax.v2.numerics import utils as numerics_utils
import jax
import jax.numpy as jnp


AbstractAqtNumerics = numerics.AqtNumerics
AbstractAqtCalibration = calibration.Calibration

AxisTiling = tiled_dot_general.AxisTiling
TilingState = tiled_dot_general.TilingState


@utils.flax_slots_kw_only_dataclass
class Quantizer:
  """Configuration of quantization of one tensor."""

  numerics: AbstractAqtNumerics = utils.static_field()
  calib_shared_axes: None | Sequence[utils.AxisIdx] | Literal["per_tensor"] = (
      utils.static_field()
  )
  scale_stop_grad: bool = utils.static_field()
  scale_dtype: None | jnp.dtype = utils.static_field(default=None)
  # noise+clip+round
  # We apply gradient of clip_and_round in bwd pass.
  calibration: None | type[AbstractAqtCalibration] = utils.static_field(
      default=None
  )
  _calibrator: None | AbstractAqtCalibration = utils.static_field(default=None)
  # TODO(yichizh): Factor out auxiliary dataclasses into a separate file.
  context: utils.Context

  # we need to speed up this initialization for the backward pass to happen
  # outside of bwd pass.
  def init_calibration(self):
    assert self._calibrator is None, "second call to self.init_calibration()"
    if self.calibration is not None:
      self._calibrator = self.calibration(dtype=self.scale_dtype)
      self._calibrator.init_calibration()

  # TODO(yichizh): Need to add type annotation back to cfg.
  def quant(
      self,
      x,
      *,
      calibration_axes: None | Sequence[utils.AxisIdx],
      tiling_state: None | TilingState = None,
  ) -> tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn]:
    """The core quantizing function."""
    qt = self.calibrate(
        x, calibration_axes=calibration_axes, tiling_state=tiling_state
    )
    qt, quant_grad = self.calculate_qvalue(x, qt)
    return qt, quant_grad

  def calibrate(
      self,
      x,
      *,
      calibration_axes: None | Sequence[utils.AxisIdx],
      tiling_state: None | TilingState = None,
  ) -> aqt_tensor.QTensor:
    """Creates incomplete QTensor with only quantization parameters.

    The tiling state is used to tile the input tensor and change the calibration
    axes accordingly. When axis is tiled, it is split into multiple tiles. Each
    tile shares the same quantization parameters like scale factor. On the other
    hand, if the axis is not tiled, the whole axis shares the same quantization
    parameters. This tiling will increase the granularity of calibration
    reducing the numeric error from quantization.

    Args:
      x: The input tensor to be calibrated.
      calibration_axes: The axes to calibrate.
      tiling_state: The tiling state of the input tensor.

    Returns:
      An incomplete QTensor with only quantization parameters.
    """

    if tiling_state:
      # Tile `x` and change calibration axes according to the tiling.
      x = tiling_state.apply(x)
      _, calibration_axes = tiling_state.to_tiled_axes_transposed(
          calibration_axes
      )

    if isinstance(self.numerics, no_numerics.NoNumerics):
      qt = aqt_tensor.QTensor(
          qvalue=x,
          scale=[],
          scale_t=None,
          bias=[],
          dequant_dtype=x.dtype,
          tiling_state=tiling_state,
      )
      return qt

    dequant_dtype = x.dtype
    # TODO(lew): We should cast earlier. xhs_q should be in cfg.xhs.dtype
    # TODO(lew): After we implement optimization to not double-quantize,
    #   what would happen if we pass fq value (xhs_q2) in residual?
    if self.calib_shared_axes == "per_tensor":
      shared_axes = list(range(x.ndim))
    else:
      shared_axes = self.calib_shared_axes or calibration_axes

    msg = (
        f"{self.calibration = } must be set if {self.numerics = } is not"
        " NoNumerics"
    )
    assert self.calibration is not None, msg
    assert self._calibrator is not None, "forgot self.init_calibration()?"
    scale, bias = self._calibrator.get_scale_and_bias(
        x, shared_axes, self.numerics, self.context
    )
    if self.scale_stop_grad:
      # TODO(lew): Does not matter in DG, because we are using custom gradient.
      #   We should take that into account somehow.
      scale = jax.lax.stop_gradient(scale)

    qt = aqt_tensor.QTensor(
        qvalue=None,
        scale=scale,
        scale_t=None,
        bias=bias,
        dequant_dtype=dequant_dtype,
        tiling_state=tiling_state,
    )
    return qt

  def calculate_qvalue(
      self, x, qt: aqt_tensor.QTensor
  ) -> tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn]:
    """Uses the quantization parameters in qt to quantize x."""
    if isinstance(self.numerics, no_numerics.NoNumerics):
      return qt, None

    # TODO: b/333984742 - make numeric as a member of QTensor, and put
    # numerics-related logics into the QTensor.
    qt = qt.quant(x)

    # TODO(lew): A logical thing would be if this call was part of
    # QTensor.quant.
    x_q, res = self.numerics.vjp_fwd(qt.qvalue, self.context)
    quant_grad = jax.tree_util.Partial(self.numerics.vjp_bwd, res)

    qt.qvalue = x_q
    return qt, quant_grad


def quantizer_make(
    n_bits: None | int | fp8_numerics.FP8Dtype,
    preserve_max_val: bool = False,
    initialize_calibration: bool = True,
    scale_stop_grad: bool = True,
    scale_dtype: None | jnp.dtype = None,
) -> Quantizer:
  """Makes Quantizer."""
  effective_numerics = numerics_utils.get_numerics(n_bits, preserve_max_val)

  if n_bits is None:
    calibration_cls = None
  else:
    calibration_cls = calibration.AbsMaxCalibration

  quantizer = Quantizer(
      numerics=effective_numerics,
      calib_shared_axes=None,
      scale_stop_grad=scale_stop_grad,
      scale_dtype=scale_dtype,
      calibration=calibration_cls,
      context=utils.Context(key=None, train_step=None),
  )
  # TODO(lew): We should try to move to to class constructor or post-init.
  # We currently need to call because bwd pass is too late for initialization.
  if initialize_calibration:
    quantizer.init_calibration()
  return quantizer


def make_fake_quant(quantizer: Quantizer, calibration_axes=None):
  def fake_quant(x):
    x_q, _ = quantizer.quant(x, calibration_axes=calibration_axes)
    return x_q.dequant()

  return fake_quant



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Resnet50 quantized model."""

import copy

from aqt.jax_legacy.jax.imagenet.configs import base_config
from aqt.jax_legacy.jax.imagenet.configs.paper.resnet50_bfloat16 import get_config as bfloat16_paper_config
from aqt.jax_legacy.jax.imagenet.configs.paper.resnet50_w4_a4_init8_dense8_auto import get_config as w4a4auto_paper_config
from aqt.jax_legacy.jax.imagenet.configs.paper.resnet50_w8_a8_auto import get_config as w8a8auto_paper_config

import ml_collections


def get_config(quant_target=base_config.QuantTarget.WEIGHTS_AND_AUTO_ACT):
  """Gets Resnet50 config for 8 bits weights and 1 bit auto activation quantization.

  conv_init and last dense layer not quantized as these are the most
  sensitive layers in the model.

  Args:
   quant_target: quantization target, of type QuantTarget.

  Returns:
   ConfigDict instance.
  """

  def set_init_bound_coeff(field):
    # input should be a class field so that the changes in this function
    # will be global to the class even without a return value
    field.cams_coeff = 0.0
    field.cams_stddev_coeff = 0.0
    field.mean_of_max_coeff = 0.0
    field.stddev_coeff = 0.0
    field.absdev_coeff = 0.0
    field.fixed_bound = 0.0
    field.granularity = "per_channel"
    field.use_old_code = False

  def reset_bound_for_convinit_dense(config):
    # reset bound haparams for conv_init and dense layers
    # use mean_of_max to automatically calculate the bound values
    set_init_bound_coeff(config.model_hparams.dense_layer.quant_act.bounds)
    config.model_hparams.dense_layer.quant_act.bounds.initial_bound = -1
    config.model_hparams.dense_layer.quant_act.bounds.mean_of_max_coeff = 1.0
    set_init_bound_coeff(config.model_hparams.conv_init.quant_act.bounds)
    config.model_hparams.conv_init.quant_act.bounds.initial_bound = -1
    config.model_hparams.conv_init.quant_act.bounds.mean_of_max_coeff = 1.0
    return config

  # create an init config which the sweep configs will be based on
  config_init = base_config.get_config(
      imagenet_type=base_config.ImagenetType.RESNET50,
      quant_target=quant_target)
  config_init.weight_prec = 8
  config_init.quant_act.prec = 1
  config_init.half_shift = True
  config_init.base_learning_rate = 2e-5
  config_init.activation_bound_start_step = 7500
  # set act function and shortcut method to each conv layer
  config_init.act_function = "none"
  config_init.shortcut_ch_shrink_method = "none"
  config_init.shortcut_ch_expand_method = "none"
  config_init.shortcut_spatial_method = "none"
  # set learning rate scheduler
  config_init.lr_scheduler.num_epochs = 250
  config_init.lr_scheduler.warmup_epochs = 5
  config_init.lr_scheduler.cooldown_epochs = 0
  config_init.lr_scheduler.scheduler = "linear"
  # -1 means no early stopping by default
  config_init.early_stop_steps = -1
  # optimizer params
  config_init.optimizer = "adam"
  config_init.adam.beta1 = 0.9
  config_init.adam.beta2 = 0.999
  # Conv_init and dense layers will have floating-point weights and acts
  config_init.model_hparams.conv_init.weight_prec = None
  config_init.model_hparams.conv_init.quant_act.prec = None
  config_init.model_hparams.dense_layer.weight_prec = None
  config_init.model_hparams.dense_layer.quant_act.prec = None
  # set all of the input distributions to "symmetric"
  config_init.model_hparams.dense_layer.quant_act.input_distribution = "symmetric"
  config_init.model_hparams.conv_init.quant_act.input_distribution = "symmetric"
  for residual_block in config_init.model_hparams.residual_blocks:
    residual_block.conv_1.quant_act.input_distribution = "symmetric"
    residual_block.conv_2.quant_act.input_distribution = "symmetric"
    residual_block.conv_3.quant_act.input_distribution = "symmetric"
    if residual_block.conv_proj is not None:
      residual_block.conv_proj.quant_act.input_distribution = "symmetric"
  # set bound hparams to all zero for activations
  # will update one of the bound hparams at a time in sweep configs
  set_init_bound_coeff(config_init.quant_act.bounds)
  # set initial bound value
  config_init.quant_act.bounds.initial_bound = 2.0
  # name of the experiment on TB
  config_init.metadata.hyper_str = "w8a1"

  # create a collection of config files for sweeping
  sweep_config = ml_collections.ConfigDict()
  configs = []

  # leaderboard configs for testing purpose
  configs.append(bfloat16_paper_config())
  configs.append(w8a8auto_paper_config())
  configs.append(w4a4auto_paper_config())

  # baseline: act_function [none, bprelu], no additional shortcuts
  for act_function in ["none", "bprelu"]:
    for fix_bound in [3.0]:
      config = copy.deepcopy(config_init)
      config.act_function = act_function
      config.quant_act.bounds.fixed_bound = fix_bound
      # reset bound haparams for conv_init and dense layers
      config = reset_bound_for_convinit_dense(config)
      config.metadata.hyper_str += f"_{act_function}_baseline"
      configs.append(config)

  # Turn on sc1, sc2, sc3
  # Sweep sc1, sc3 both with different methods
  for act_function in ["bprelu"]:
    for fix_bound in [3.0]:
      for shortcut_shrink_method in ["consecutive"]:
        for shortcut_expand_method in ["zeropad"]:
          config = copy.deepcopy(config_init)
          config.act_function = act_function
          config.quant_act.bounds.fixed_bound = fix_bound
          config.shortcut_ch_shrink_method = shortcut_shrink_method
          config.shortcut_ch_expand_method = shortcut_expand_method
          config.shortcut_spatial_method = "max_pool"
          # reset bound haparams for conv_init and dense layers
          config = reset_bound_for_convinit_dense(config)
          config.metadata.hyper_str += f"_{act_function}_sc123_shrink_{shortcut_shrink_method}_expand_{shortcut_expand_method}"
          configs.append(config)

  sweep_config.configs = configs
  return sweep_config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""DotGeneralQuantizer for GPTQ."""

# Lingo in this file:
#
# - lhs(rhs) - left(right) hand side of a binary operation
# - ca - contraction axes
# - hinv - Inverse of the hessian

from typing import Sequence
from aqt.jax.v2 import aqt_dot_general
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import int_numerics
import flax.linen as nn
import jax
from jax import numpy as jnp
import numpy as np
from rigl.projects.bigsparse import sharded_gptc


def _get_quant_mode(context: None | utils.Context) -> utils.QuantMode:
  return utils.QuantMode.TRAIN if context is None else context.quant_mode


def _get_divisible_blocksize(dim: int, blocksize_top: int) -> int:
  """Returns the blocksize which could divide the given dimension.

  Returns the largest blocksize which is smaller than or equal to
  blocksize_top.

  Args:
    dim: Target dimension which should be divided.
    blocksize_top: Maximal dimension size.

  Returns:
    The largest blocksize which is smaller than blocksize_top, and which could
    divide the given dim.
  """
  blocksize = blocksize_top
  while dim % blocksize != 0:
    blocksize -= 1

  return blocksize


def _reshape_kernel_for_gptq(
    kernel: jnp.ndarray,
    ca: Sequence[utils.AxisIdx],
    sharding_axes: None | str,
    act_order: bool,
    perm: None | Sequence[utils.AxisIdx],
    blocksize: int,
) -> tuple[jnp.ndarray, Sequence[int]]:
  """Reshapes kernel to (features / blocksize, blocksize, -1) and potentially reshard."""
  # TODO(dhchoi): Consider using kernel[XXX:] form, check its efficiency.
  for new_axis_idx, original_axis_idx in enumerate(sorted(ca)):
    kernel = jnp.moveaxis(kernel, original_axis_idx, new_axis_idx)

  kernel_feature_grouped_shape = kernel.shape
  kernel = kernel.reshape((np.prod(kernel.shape[: len(ca)]), -1))
  if act_order:
    assert perm is not None
    kernel = kernel[perm, :]

  kernel = sharded_gptc.maybe_shard(
      kernel,
      jax.sharding.PartitionSpec(sharding_axes, None),
      sharding_axes,
  )
  kernel = jnp.float32(kernel)  # convert to higher precision after sharding

  return (
      kernel.reshape((-1, blocksize, kernel.shape[-1])),
      kernel_feature_grouped_shape,
  )


def _recover_kernel_from_gptq_result(
    kernel: jnp.ndarray,
    ca: Sequence[utils.AxisIdx],
    sharding_axes: None | str,
    act_order: bool,
    perm: None | Sequence[utils.AxisIdx],
    kernel_dtype: jnp.dtype,
    kernel_feature_grouped_shape: Sequence[int]
) -> jnp.ndarray:
  """Recovers original kernel shape."""
  kernel = kernel.astype(kernel_dtype)
  if act_order:
    assert perm is not None
    invperm = jnp.argsort(perm)  # pytype: disable=wrong-arg-types
    kernel = kernel.reshape((-1, kernel.shape[-1]))
    kernel = kernel[invperm, :]
  kernel = kernel.reshape(kernel_feature_grouped_shape)

  for original_axis_idx, new_axis_idx in list(enumerate(ca))[::-1]:
    kernel = jnp.moveaxis(kernel, original_axis_idx, new_axis_idx)

  kernel = sharded_gptc.maybe_shard(kernel, None, sharding_axes)
  return kernel


def _init_hinv_for_calibration(inputs, perc_damp=0.01):
  """Initializes hinv with a damping term."""
  features = np.prod(inputs.shape[2:])
  diag = jnp.mean(jnp.float32(inputs) ** 2, (0, 1)).flatten()
  damp = perc_damp * jnp.mean(diag)
  return jnp.diag(jnp.ones(features) / damp)


class GptqHinvCollector(nn.Module):
  """GPTQ hinv collector module.

  Written as a separated module since if we make the GptqDotGeneralQuantizer as
  a module, we cannot inject its instance (module cannot be initialized outside
  of flax.init / flax.apply). If we inject the type directly, then we cannot
  use its member utility functions (ex. swap_lhs_and_rhs) during configuration.
  """
  quant_collection: str

  sharding_axes: None | str

  # Percentage of damping during hinv initialization.
  perc_damp: float = 0.01

  # If act_order is set, weights with large corresponding activation values are
  # updated first.
  act_order: bool = False

  @nn.compact
  def __call__(
      self,
      x: jnp.ndarray,
      ca: Sequence[utils.AxisIdx],
      quant_mode: utils.QuantMode,
  ) -> tuple[jnp.ndarray, None | jnp.ndarray]:
    """Collects Inverse of the hessian.

    Args:
      x: Activation using which the hessian will be collected.
      ca: Contracting axis.
      quant_mode: Quantization mode.
    Returns:
      Collected hessian inverse.
    """
    # The GPTQ assumes that 2+ dimensions are all the reducing dimension.
    # We should reshape so that the shape is something like BS X S X D, where
    # D is the contracting dimension.
    transpose = [axis for axis in range(x.ndim) if axis not in ca] + list(ca)
    x = x.transpose(transpose)
    x = x.reshape((-1, np.prod(x.shape[-len(ca) :])))
    x = sharded_gptc.maybe_shard(
        x,
        jax.sharding.PartitionSpec(None, self.sharding_axes),
        self.sharding_axes,
    )

    blocksize = _get_divisible_blocksize(x.shape[0], sharded_gptc.BLOCKSIZE)
    x = x.reshape((-1, blocksize, x.shape[-1]))

    is_initializing = not self.has_variable(
        self.quant_collection, "num_calibrated_batches"
    )
    num_calibrated_batches = self.variable(
        self.quant_collection,
        "num_calibrated_batches",
        jnp.zeros,
        (),
        jnp.int32,
    )

    hinv = None
    if not is_initializing:
      # Collect hinv from the previous calibration step.
      hinv = self.get_variable(self.quant_collection, "collected_hinv")
      hinv = jax.lax.cond(
          num_calibrated_batches.value == 0,
          lambda: _init_hinv_for_calibration(x, self.perc_damp),
          lambda: hinv
      )

    # This function, when called with hinv = None, calculates
    # (D + (X^T @ X) / bs)^-1, where bs = batch size and D is damping term.
    # When called with hinv != None, it aggregates the impact of additional
    # (X^T @ X) / bs on the hinv. This is possible since the the impact of each
    # data instance is accumulated using Woodbury formula on the hinv.
    hinv, perm = sharded_gptc.compute_hinv(
        x,
        perc_damp=self.perc_damp,
        sharding_axes=self.sharding_axes,
        act_order=self.act_order,
        blocksize=blocksize,
        hinv=hinv,
    )

    collected_hinv = self.variable(
        self.quant_collection,
        "collected_hinv",
        jnp.zeros,
        hinv.shape,
        hinv.dtype,
    )

    if quant_mode == utils.QuantMode.CALIBRATE and not is_initializing:
      collected_hinv.value = hinv
      num_calibrated_batches.value = num_calibrated_batches.value + 1

    # The collected hinv will be (D + (C^T @ C) / bs)^-1, where C is the whole
    # calibration data with instance number = bs * bn (bn is batch num).
    # We multiply the collected hinv by bn. By doing so, the result will be
    # bn * (D + (C^T @ C) / bs)^-1 = (D / bn + (C^T @ C) / (bs * bn))^-1).
    # Althogh the damping term is divided by bn, the values calculated using the
    # calibration data remains the same with the case when the whole calibration
    # data is used at once to calculate the hinv. In addition, we can manually
    # configure perc_damp (divide it by the number of calibration batches) to
    # approximate it more precise.
    return collected_hinv.value * num_calibrated_batches.value, perm


@utils.flax_slots_kw_only_dataclass
class GptqDotGeneralQuantizer(aqt_dot_general.DefaultDotGeneralQuantizer):
  """GPTQ dot_general quantizer."""

  sharding_axes: None | str = utils.static_field()

  quant_collection: str = utils.static_field()

  # Percentage of damping during hinv initialization.
  perc_damp: float = utils.static_field(default=0.01)

  # If act_order is set, weights with large corresponding activation values are
  # updated first.
  act_order: bool = utils.static_field(default=False)

  # Boolean flag to see which side of argument is the kernel.
  is_rhs_kernel: bool = utils.static_field(default=True)

  def calibrate(
      self,
      lhs: jax.Array,
      rhs: jax.Array,
      dimension_numbers: None | jax.lax.DotDimensionNumbers,
      lhs_mode: None | aqt_dot_general.CalibrationMode,
      rhs_mode: None | aqt_dot_general.CalibrationMode,
  ) -> tuple[
      tuple[jax.Array, aqt_tensor.QTensor], tuple[jax.Array, aqt_tensor.QTensor]
  ]:  # pylint: disable=g-doc-args
    """GPTQ calibration.

    Majority of the codes are copied from sharded_gptc.gptc_skeleton. we copied
    the code snippet to enable the collection of hinv statistics.

    Returns:
      Updated weight and incomplete QTensor for lhs and rhs.
    """
    if dimension_numbers is None:
      lhs_calib, rhs_calib = None, None
    else:
      (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers

      def _get_calibration_axes(
          mode: aqt_dot_general.CalibrationMode,
          ndim: int,
          ca: Sequence[utils.AxisIdx],
          ba: Sequence[utils.AxisIdx],
      ) -> Sequence[utils.AxisIdx]:
        """Computes calibration axes for the given Tensor."""
        match mode:
          case aqt_dot_general.CalibrationMode.REMAINING_AXIS:
            calibration_axes = utils.get_remaining_axes(ndim, ca, ba)
          case aqt_dot_general.CalibrationMode.CONTRACTING_AXIS:
            calibration_axes = ca
          case _:
            raise ValueError(f"Unknown calibration mode: {mode}")
        return calibration_axes

      lhs_calib = _get_calibration_axes(lhs_mode, lhs.ndim, lhs_ca, lhs_ba)
      rhs_calib = _get_calibration_axes(rhs_mode, rhs.ndim, rhs_ca, rhs_ba)
    lhs_quantization_info = (lhs, lhs_calib)
    rhs_quantization_info = (rhs, rhs_calib)

    # TODO(lew): The block of code above is copied from
    # aqt_dot_general.DefaultDotGeneralQuantizer.calibrate.
    # [lr]hs_quantization_info are the arguments that were passed before.
    # We should refactor it to make it more readable.
    # TODO(lew): We should also consider refactoring the code below to make it
    # more readable.
    # Perhaps we can drop "The GPTQ assumes that 2+ dimensions" assumption too.

    # Follow the quantization mode and num_bits of the kernel.
    if self.is_rhs_kernel:
      quant_mode = _get_quant_mode(self.rhs.context)
      assert isinstance(self.rhs.numerics, int_numerics.IntSymmetric)
      num_bits = self.rhs.numerics.bits
    else:
      quant_mode = _get_quant_mode(self.lhs.context)
      assert isinstance(self.lhs.numerics, int_numerics.IntSymmetric)
      num_bits = self.lhs.numerics.bits

    if quant_mode == utils.QuantMode.TRAIN:
      # During training, we should not allow collecting hinvs and updating
      # weights using it.
      return super(GptqDotGeneralQuantizer, self).calibrate(  # pytype: disable=attribute-error
          lhs, rhs, dimension_numbers, lhs_mode, rhs_mode
      )

    if not self.is_rhs_kernel:
      # Swap so that rhs is kernel.
      lhs_quantization_info, rhs_quantization_info = (
          rhs_quantization_info,
          lhs_quantization_info,
      )

    lhs, lhs_ca = lhs_quantization_info
    rhs, rhs_ca = rhs_quantization_info

    hinv_collector = GptqHinvCollector(
        quant_collection=self.quant_collection,
        sharding_axes=self.sharding_axes,
        perc_damp=self.perc_damp,
        act_order=self.act_order,
    )
    hinv, perm = hinv_collector(lhs, lhs_ca, quant_mode)

    # Cholesky decomposition.
    blocksize = _get_divisible_blocksize(hinv.shape[0], sharded_gptc.BLOCKSIZE)
    hinv = sharded_gptc.cholesky(
        hinv, sharding_axes=self.sharding_axes, blocksize=blocksize
    )

    # Reshape rhs to (features, -1) and potentially reshard
    rhs_dtype = rhs.dtype

    rhs, rhs_feature_grouped_shape = _reshape_kernel_for_gptq(
        rhs, rhs_ca, self.sharding_axes, self.act_order, perm, blocksize
    )
    hinv = hinv.reshape((-1, blocksize, hinv.shape[-1]))

    def find_qparams(w, solve=None):
      # Fake-Quant calculator used by the GPTQ.
      # W is expected to have the shape C X D, while C is the contracting.
      del solve

      # Add a smoothing term to prevent it from breaking by having max value 0
      # in some block. (Ex: when w is identity + subchannel)
      return sharded_gptc.absmax_params(w + 1e-7, bits=num_bits)

    # The GPTQ:
    # 1. Fake-quant the weight.
    # 2. Calculate quantization error of the weight.
    # 3. Outer-product hessian inverse and quantization error; update W using
    #    the product value.
    # 4. Weight is updated sequentially.
    rhs, _ = sharded_gptc.gptq(rhs, hinv, find_qparams)

    # Recover original rhs shape.
    rhs = _recover_kernel_from_gptq_result(
        rhs,
        rhs_ca,
        self.sharding_axes,
        self.act_order,
        perm,
        rhs_dtype,
        rhs_feature_grouped_shape,
    )

    # Restore lhs and rhs position.
    if not self.is_rhs_kernel:
      lhs, rhs = rhs, lhs
      lhs_ca, rhs_ca = rhs_ca, lhs_ca

    # Retrieve the scales using the updated lhs and rhs.
    return super(GptqDotGeneralQuantizer, self).calibrate(  # pytype: disable=attribute-error
        lhs, rhs, dimension_numbers, lhs_mode, rhs_mode
    )

  def swap_lhs_and_rhs(self) -> None:
    """Swaps lhs and rhs configuration."""
    self.lhs, self.rhs = self.rhs, self.lhs
    self.is_rhs_kernel = not self.is_rhs_kernel



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A very simple non-quantized execution sample of a CNN model."""

from collections.abc import Sequence

from absl import app
from aqt.jax.v2.examples.cnn import cnn_model
from aqt.jax.v2.examples.cnn import model_utils
import jax


def run(train_ds: dict[str, jax.Array], test_ds: dict[str, jax.Array]) -> None:
  """Extracted main function for unit testing."""
  # 1. Train
  state = model_utils.create_train_state(
      rng=jax.random.key(0),
      train_model=cnn_model.CNN(True),
      eval_model=cnn_model.CNN(False),
  )
  state = model_utils.train_and_evaluate(
      num_epochs=1,
      workdir='/tmp/aqt_mnist_example',
      train_ds=train_ds,
      test_ds=test_ds,
      state=state,
  )

  # 2. Serve and test
  loss = model_utils.serve(
      serve_model=cnn_model.CNN(False),
      model_vars=state.model_vars,
      test_ds=test_ds,
  )
  print('serve loss on sample ds: {}'.format(loss))


def main(argv: Sequence[str]) -> None:
  del argv

  train_ds, test_ds = cnn_model.get_datasets()
  run(train_ds, test_ds)


if __name__ == '__main__':
  app.run(main)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for training_hparams.py."""
import copy
import dataclasses
import json

from absl import flags
from absl.testing import absltest
from absl.testing import parameterized

from aqt.jax_legacy.jax.wmt_mlperf import train
from aqt.jax_legacy.jax.wmt_mlperf import training_hparams
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs import base_config
from aqt.jax_legacy.utils import hparams_utils
import jax


FLAGS = flags.FLAGS


class HParamsTest(parameterized.TestCase):

  def test_dynamic_hparams(self):
    # Check that a JITed function sees mutations to hparams arguments. We check
    # that mutating the learning rate (arbitrary choice) of an hparams instance
    # causes Jax to appropriately recompile. This test is useful since Jax
    # caches static arguments to JITed functions and we want to make sure that
    # Jax's recommended workflow of deep-copying an hparams instance that is
    # modified between training steps works correctly.
    def return_learning_rate(hparams):
      return hparams.learning_rate_schedule.base_learning_rate

    return_learning_rate_jit = jax.jit(
        return_learning_rate, static_argnums=(0,))
    hparams = hparams_utils.load_dataclass_from_config_dict(
        training_hparams.TrainingHParams,
        base_config.get_config(
            n_layers=3,
            quant_target=base_config.QuantTarget.WEIGHTS_AND_AUTO_ACTS))
    hparams.learning_rate_schedule.base_learning_rate = 1.0
    learning_rate = return_learning_rate_jit(hparams)
    self.assertEqual(learning_rate, 1.0)
    hparams = copy.deepcopy(hparams)
    hparams.learning_rate_schedule.base_learning_rate = 2.0
    mutated_learning_rate = return_learning_rate_jit(hparams)
    self.assertEqual(mutated_learning_rate, 2.0)


  def test_convert_lists_to_tuples(self):
    test_input = {
        'a': [
            {  #
                'b': [1, 2]
            },  #
            {
                'b': [3, 4]
            }
        ],
        'c': 'd',
        'e': [],
        'f': (1,)
    }
    test_output = hparams_utils._convert_lists_to_tuples(test_input)
    self.assertEqual(
        test_output,
        {
            'a': (
                {  #
                    'b': (1, 2)
                },  #
                {
                    'b': (3, 4)
                }  #
            ),
            'c': 'd',
            'e': (),
            'f': (1,)
        })
    test_input = [1, 2, 3]
    test_output = hparams_utils._convert_lists_to_tuples(test_input)
    self.assertEqual(test_output, (1, 2, 3))
    test_input = 1
    test_output = hparams_utils._convert_lists_to_tuples(test_input)
    self.assertEqual(test_output, 1)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized conv_general."""

# Lingo in this file:
#
# - lhs(rhs) - left(right) hand side of a binary operation
# - ca - contraction axes
# - ba - batch axes
# - ra - remaining axes

# pylint: disable=protected-access

from typing import Tuple
from aqt.jax.v2 import aqt_dot_general
from aqt.jax.v2 import aqt_tensor
import jax
from jax import lax
import jax.numpy as jnp


def make_conv_general_dilated_with_qt(cfg: aqt_dot_general.DotGeneralRaw):
  """Makes quantized lax.make_conv_general_dilated replacement which returns QTensors."""
  # TODO(lew): Either rename DotGeneralConfig or make a conv-specific cfg.
  assert cfg is not None, "Missing config for make_conv_general_dilated"

  def my_conv_general_dilated(
      lhs,
      rhs,
      window_strides,
      padding,
      lhs_qt: None | aqt_tensor.QTensor = None,
      rhs_qt: None | aqt_tensor.QTensor = None,
      lhs_dilation=None,
      rhs_dilation=None,
      dimension_numbers=None,
      feature_group_count=1,
      batch_group_count=1,
      precision=None,
      preferred_element_type=None,
  ) -> Tuple[jax.Array, tuple[aqt_tensor.QTensor, aqt_tensor.QTensor]]:
    msg1 = """
To simplify the code, we currently assume a Flax-particular layout of the data.
This makes sense, because this is the main use-case of this function.
However if there is any other use, we will drop that assumption."""
    rank = len(lhs.shape)
    assert len(rhs.shape) == rank
    assert dimension_numbers is not None, msg1
    assert dimension_numbers.lhs_spec[0:2] == (0, rank - 1), msg1
    assert dimension_numbers.rhs_spec[0:2] == (rank - 1, rank - 2), msg1
    assert dimension_numbers.out_spec[0:2] == (0, rank - 1), msg1
    # In Flax, lhs is the inputs, rhs is the kernel.
    # lhs layout is B, spatials..., Ci
    # rhs layout is: spatials..., Ci, Co
    # out layout is: B, spatials..., Co
    #
    # we need to share these axes: lhs[1:] , rhs[:-1]
    # we have a scale/invscale per: lhs[0] / out[0] and rhs[-1] / out[-1]

    # Flax assumptions.
    msg = (
        "Convolution formula does not follow flax assumption. This could be"
        " because spatial dimensions was incorrectly set during DotGeneralRaw"
        " creation. Please double check the parameter in"
        " #conv_general_dilated_make()."
    )
    # TODO(lew): Perhaps we should rely only on passing  passing calib shared
    # axes value instead of setting it in config. (we pass None below)
    cfg.dg_quantizer.assert_calib_shared_axes_value(
        list(range(1, rank)), list(range(0, rank - 1)), msg
    )
    (lhs, lhs_incomplete_qt), (rhs, rhs_incomplete_qt) = (
        cfg.dg_quantizer.calibrate(lhs, rhs, None, None, None)
    )
    if lhs_qt is not None and not lhs_qt.is_full():
      # Incomplete QTensor is provided as lhs_qt.
      lhs_incomplete_qt = lhs_qt
      lhs_qt = None

    if rhs_qt is not None and not rhs_qt.is_full():
      # Incomplete QTensor is provided as rhs_qt.
      rhs_incomplete_qt = rhs_qt
      rhs_qt = None

    lhs_quantized, rhs_quantized = cfg.dg_quantizer.calculate_qvalue(
        lhs, lhs_incomplete_qt, rhs, rhs_incomplete_qt
    )
    lhs_qt_calculated, _ = lhs_quantized
    rhs_qt_calculated, _ = rhs_quantized

    lhs_qt = lhs_qt if lhs_qt is not None else lhs_qt_calculated
    rhs_qt = rhs_qt if rhs_qt is not None else rhs_qt_calculated

    out = lax.conv_general_dilated(
        lhs=lhs_qt.qvalue_astype(lhs.dtype).qvalue,
        rhs=rhs_qt.qvalue_astype(rhs.dtype).qvalue,
        window_strides=window_strides,
        padding=padding,
        lhs_dilation=lhs_dilation,
        rhs_dilation=rhs_dilation,
        dimension_numbers=dimension_numbers,
        feature_group_count=feature_group_count,
        batch_group_count=batch_group_count,
        precision=precision,
        preferred_element_type=preferred_element_type,
    )

    # It seems lucky that original scale has shape suitable for output
    # scaling without any transposition.
    out = aqt_tensor.QTensor(
        qvalue=out,
        scale=[],
        scale_t=None,
        bias=[],
        dequant_dtype=jnp.promote_types(lhs, rhs),
    )
    assert out.scale is not None  # pytype help
    out.scale.extend(lhs_qt.scale)
    out.scale.extend(rhs_qt.scale)
    out = out.dequant()

    # # Future scale granularity optimization.
    # In 1x1 conv, each pixel (spatial location) can have different scales
    # in 1xN (rows x columns) conv each row can have different scale, but
    # columns need to share the scales ,  because we are adding pixels across.
    #
    # For patch convs we could have separate scales per patch.
    # We don't do that optimization, because there is a  Flax op: ConvLocal
    # using lax.conv_general_dilated_local which uses lax.dot_general.
    #
    # Dilations: If a dilation of LHS is bigger than the total spatial size of
    # RHS, we could use separe (per LHS pixel) scales.
    # The same applies to dilated RHS.
    # We don't do that optimization yet.
    #
    # We can have different scales across different groups.
    # This applies to both feature and batch.
    return out, (lhs_qt, rhs_qt)

  return my_conv_general_dilated


def make_conv_general_dilated(cfg: aqt_dot_general.DotGeneralRaw):
  """Makes quantized lax.make_conv_general_dilated replacement."""
  assert cfg is not None, "Missing config for make_conv_general_dilated"

  def my_conv_general_dilated(
      lhs,
      rhs,
      window_strides,
      padding,
      lhs_dilation=None,
      rhs_dilation=None,
      dimension_numbers=None,
      feature_group_count=1,
      batch_group_count=1,
      precision=None,
      preferred_element_type=None,
  ) -> jax.Array:
    my_conv_general_dilated_with_qt = make_conv_general_dilated_with_qt(cfg)

    out, _ = my_conv_general_dilated_with_qt(
        lhs,
        rhs,
        window_strides,
        padding,
        lhs_qt=None,
        rhs_qt=None,
        lhs_dilation=lhs_dilation,
        rhs_dilation=rhs_dilation,
        dimension_numbers=dimension_numbers,
        feature_group_count=feature_group_count,
        batch_group_count=batch_group_count,
        precision=precision,
        preferred_element_type=preferred_element_type,
    )
    return out

  return my_conv_general_dilated


def conv_general_dilated_make(
    spatial_dimensions: int,
    lhs_bits: None | int = None,
    rhs_bits: None | int = None,
    initialize_calibration: bool = True,
) -> aqt_dot_general.DotGeneralRaw:
  """Create quantization config conv_general_dilated.

  Args:
    spatial_dimensions: The number of dimensions of the base area that the
      convolutional window moves across.
    lhs_bits: The precision for quantization for lhs
    rhs_bits: The precision for quantization for rhs
    initialize_calibration: If set, calibrators are initialized. Set False if
      you want to run static range quantization.

  Returns:
    DotGeneralRaw object to be injected into nn.Conv as conv_general_dilated.
  """
  config = aqt_dot_general.dot_general_raw_make(
      lhs_bits, rhs_bits, initialize_calibration=initialize_calibration)
  # Hardcoding flax assumptions.
  lhs_calib_shared_axes = (
      list(range(1, spatial_dimensions + 2)) if config.lhs else None
  )
  rhs_calib_shared_axes = (
      list(range(0, spatial_dimensions + 2 - 1)) if config.rhs else None
  )

  assert isinstance(
      config.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  config.dg_quantizer.lhs.calib_shared_axes = lhs_calib_shared_axes
  config.dg_quantizer.rhs.calib_shared_axes = rhs_calib_shared_axes

  return config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""AQT intercept methods."""

import abc
from typing import Any, Callable, Mapping, Sequence

from aqt.jax.v2.flax import aqt_flax
import flax
from flax import linen as nn
import jax
from jax import lax


_Args = Sequence[Any]
_Kwargs = Mapping[str, Any]
_NextGetter = Callable[..., Any]
_Interceptor = Callable[
    [_NextGetter, _Args, _Kwargs, flax.linen.module.InterceptorContext], Any
]


class _DotGeneralScope:
  def __init__(self, dot_general):
    self.dot_general = dot_general

  def __enter__(self):
    self.orignal_dot_general = lax.dot_general
    lax.dot_general = self.dot_general
    return self

  def __exit__(self, exception_type, exception_value, traceback):
    lax.dot_general = self.orignal_dot_general


class DotGeneralGeneratorBase(metaclass=abc.ABCMeta):
  """Abstract class to generate dot general.

  Attributes:
    dot_general: Saved the dot general for the generator.
  """

  @abc.abstractmethod
  def __call__(
      self,
      next_f: _NextGetter,
      args: _Args,
      kwargs: _Kwargs,
      context: flax.linen.module.InterceptorContext):
    """Generate dot general method by intercepted arguments.

    Args:
      next_f: A function wrapped by the interceptor.
      args: Arguments to the function call.
      kwargs: Keyword arguments to the function call.
      context: Context in which the interceptor is applied.

    Returns:
      The dot general method for the intercepted context.
    """
    raise NotImplementedError


class DotGeneralGenerator(DotGeneralGeneratorBase):
  """Generate saved dot general."""

  def __init__(self, dot_general: Any):
    """Initializes dot general generator with given dot general method.

    Args:
      dot_general: Saved the dot general.
    """
    self.dot_general = dot_general

  def __call__(
      self,
      next_f: _NextGetter,
      args: _Args,
      kwargs: _Kwargs,
      context: flax.linen.module.InterceptorContext) -> Any:
    """Return the dot general method passed on initializes.

    Args:
      next_f: A function wrapped by the interceptor.
      args: Arguments to the function call.
      kwargs: Keyword arguments to the function call.
      context: Context in which the interceptor is applied.

    Returns:
      The saved dot general method.
    """
    del next_f, args, kwargs, context
    return self.dot_general


# TODO(kimjaehong): This generator filtered intercept state by module class to
# support flax examples. To support enhanced transformer based LLMs, we need to
# add filters by module name to distinguish between attention and feedforward.
class DotGeneralGeneratorByModule(DotGeneralGenerator):
  """Abstract class to generate dot general by module."""

  def __call__(
      self,
      next_f: _NextGetter,
      args: _Args,
      kwargs: _Kwargs,
      context: flax.linen.module.InterceptorContext):
    """Return the dot general method by module on the context.

    Args:
      next_f: A function wrapped by the interceptor.
      args: Arguments to the function call.
      kwargs: Keyword arguments to the function call.
      context: Context in which the interceptor is applied.

    Returns:
      The dot general method for the module on the context.
    """
    ret = self.generate_by_module(context.module)
    if ret is None:
      ret = super().__call__(next_f, args, kwargs, context)
    return ret

  @abc.abstractmethod
  def generate_by_module(self, module: nn.Module):
    """Generate dot general method by the intercepted module object.

    Args:
      module: An intercepted module from the intercept methods.

    Returns:
      The dot general method for the intercepted module.
    """
    raise NotImplementedError


def intercept_methods_replace_dot_general(
    dot_general_generator: DotGeneralGeneratorBase = DotGeneralGenerator(
        lax.dot_general
    ),
):
  # pylint: disable=g-doc-return-or-yield
  """Flax intercept method wrapper to replace dot general.

  Args:
    dot_general_generator: The dot general generator by the intercepted args.
  """
  dot_general_cache = {}

  def dot_general_interceptor(
      next_f: _NextGetter,
      args: _Args,
      kwargs: _Kwargs,
      context: flax.linen.module.InterceptorContext,
  ) -> Any:
    """Dot general interceptor that replace dot general by module.

    Args:
      next_f: A function wrapped by the interceptor.
      args: Arguments to the function call.
      kwargs: Keyword arguments to the function call.
      context: Context in which the interceptor is applied.

    Returns:
      The result of original function, with dot general scope.
    """
    layer_path = '/'.join(context.module.scope.path)
    if layer_path not in dot_general_cache:
      if not context.module._initialization_allowed:  # pylint: disable=protected-access
        raise ValueError(
            f'The tracing function {context.method_name} is not compact, and'
            f' the module {layer_path} does not have setup. Cannot initialize'
            ' dot general module...')
      dot_general_cache[layer_path] = dot_general_generator(
          next_f, args, kwargs, context)

    with _DotGeneralScope(dot_general_cache[layer_path]):
      out = next_f(*args, **kwargs)
    return out

  return nn.intercept_methods(dot_general_interceptor)


class AqtDotGeneralGenerator(DotGeneralGeneratorByModule):
  """Generate AQT dot general by module."""

  def __init__(self, *args, **kwargs):
    super().__init__(jax.lax.dot_general)
    self.aqt_args = args
    self.aqt_kwargs = kwargs

  def generate_by_module(self, module: nn.Module):
    """Generate AQT dot general method by the intercepted module object.

    Args:
      module: An intercepted module from the intercept methods.

    Returns:
      The AQT dot general method for the intercepted module.
    """
    # Replace `jax.lax.dot_general` under the all dense layer methods.
    if isinstance(module, nn.Dense):
      return aqt_flax.AqtDotGeneral(*self.aqt_args, **self.aqt_kwargs)
    return None


def intercept_methods(*args, **kwargs):
  return intercept_methods_replace_dot_general(
      AqtDotGeneralGenerator(*args, **kwargs))


def intercept_wrapper(func, *aqt_args, **aqt_kwargs):
  def wrapper(*args, **kwargs):
    with intercept_methods(*aqt_args, **aqt_kwargs):
      return func(*args, **kwargs)
  return wrapper




# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.jax.hlo_utils."""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax import hlo_utils
from flax import linen as nn
from jax import random
import jax.numpy as jnp


class HloUtilsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name='one_add',
          fn=lambda x: x + 1,
          fn_args=[1],
          ops_regex=r'add',
          exp_count=1,
      ),
      dict(
          testcase_name='two_adds',
          fn=lambda x, y: x + y + 1,
          fn_args=[1, 2],
          ops_regex=r'add',
          exp_count=2,
      ),
      dict(
          testcase_name='one_mult',
          fn=lambda x, y: x * y,
          fn_args=[2, 3],
          ops_regex=r'multiply',
          exp_count=1,
      ),
  )
  def test_load_hlo_proto_from_jax_fn_and_count_ops(self, fn, fn_args,
                                                    ops_regex, exp_count):
    hlo_proto = hlo_utils.load_hlo_proto_from_jax_fn(fn, *fn_args)
    count = hlo_utils.count_ops_in_hlo_proto(hlo_proto, ops_regex=ops_regex)
    self.assertEqual(count, exp_count)

  class TestModelWith2DenseLayers(nn.Module):
    """Test model with two Dense layers."""

    @nn.compact
    def __call__(self, inputs, dtype=jnp.float32):
      x = nn.linear.Dense(features=2)(inputs)
      output = nn.linear.Dense(features=3)(x)
      return output

  def test_load_hlo_proto_from_model_and_count_ops(self):
    input_shapes = [(1, 2)]
    # with nn.stateful() as init_state:
    test_model = self.TestModelWith2DenseLayers()
    init_state = test_model.init(
        random.PRNGKey(0), *[jnp.ones(shape) for shape in input_shapes])

    hlo_proto = hlo_utils.load_hlo_proto_from_model(test_model, init_state,
                                                    input_shapes)
    count = hlo_utils.count_ops_in_hlo_proto(hlo_proto, ops_regex=r'dot')
    self.assertEqual(count, 2)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import aqt_quantizer
import aqt.jax.v2.aqt_conv_general as aqt_conv
import flax.linen.linear as fl
import jax
import jax.numpy as jnp


def rand_unif(shape, maxval, seed, dtype=jnp.float32):
  key = jax.random.PRNGKey(seed)
  return jax.random.uniform(
      key=key, shape=shape, minval=-maxval, maxval=maxval, dtype=dtype
  )


def _apply_po2_scale(quantizer):
  if quantizer.calibration is None:
    return

  calibration_cls = quantizer.calibration
  # TODO(lew): Remove partial inspection wherever possible.
  # Partial inspection is needed because the current implementation of delayed
  # calibration initialization requires the configuration to be set via
  # functools.partial.
  keywords = {}
  if isinstance(calibration_cls, functools.partial):
    keywords = calibration_cls.keywords
    calibration_cls = calibration_cls.func
  keywords.update(po2_scale=True)
  quantizer.calibration = functools.partial(calibration_cls, **keywords)


class AqtConvGeneralTest(parameterized.TestCase):

  @parameterized.parameters([
      (1, 1),
      (1, 2),
      (2, 1),
      (2, 2),
      (8, 8),
      (None, 8),
      (8, None),
      (None, None),
  ])
  def test_conv_general_dilated(
      self,
      lhs_bits,
      rhs_bits,
      lhs_maxval=10.0,
      rhs_maxval=20.0,
      seed=0,
  ):
    dg_raw_conv = aqt_conv.conv_general_dilated_make(
        2, lhs_bits, rhs_bits, initialize_calibration=False
    )
    # Power-of-2 scales allow FQ and AQT to be exactly the same.
    dg_quantizer = dg_raw_conv.dg_quantizer
    if dg_raw_conv.lhs:
      _apply_po2_scale(dg_quantizer.lhs)
      dg_quantizer.lhs.init_calibration()
    if dg_raw_conv.rhs:
      _apply_po2_scale(dg_quantizer.rhs)
      dg_quantizer.rhs.init_calibration()

    batch_n = 10
    contr_n = 20
    feature_n = 30
    lhs = rand_unif((batch_n, 4, 5, contr_n), lhs_maxval, seed)
    rhs = rand_unif((3, 3, contr_n, feature_n), rhs_maxval, seed + 1)

    lax_conv = jax.lax.conv_general_dilated
    aqt_conv_fn = aqt_conv.make_conv_general_dilated(dg_raw_conv)
    kwargs = {
        "window_strides": (1, 1),
        "padding": "SAME",
        "dimension_numbers": fl._conv_dimension_numbers(lhs.shape),
    }
    lhs_fq = aqt_quantizer.make_fake_quant(dg_raw_conv.dg_quantizer.lhs)(lhs)
    rhs_fq = aqt_quantizer.make_fake_quant(dg_raw_conv.dg_quantizer.rhs)(rhs)
    prod_fq = lax_conv(lhs_fq, rhs_fq, **kwargs)
    prod_aqt = aqt_conv_fn(lhs, rhs, **kwargs)
    assert (prod_aqt == prod_fq).all()

  @parameterized.parameters([
      (1, 1),
      (1, 2),
      (2, 1),
      (2, 2),
      (8, 8),
      (None, 8),
      (8, None),
      (None, None),
  ])
  def test_conv_general_dilated_quantized(
      self,
      lhs_bits,
      rhs_bits,
      lhs_maxval=10.0,
      rhs_maxval=20.0,
      seed=0,
  ):
    """Check that passing quantized lhs/rhs to aqt_conv_fn works."""
    dg_raw_conv = aqt_conv.conv_general_dilated_make(
        2, lhs_bits, rhs_bits, initialize_calibration=False
    )
    # Power-of-2 scales allow FQ and AQT to be exactly the same.
    dg_quantizer = dg_raw_conv.dg_quantizer
    if dg_raw_conv.lhs:
      _apply_po2_scale(dg_quantizer.lhs)
      dg_quantizer.lhs.init_calibration()
    if dg_raw_conv.rhs:
      _apply_po2_scale(dg_quantizer.rhs)
      dg_quantizer.rhs.init_calibration()

    batch_n = 10
    contr_n = 20
    feature_n = 30
    lhs = rand_unif((batch_n, 4, 5, contr_n), lhs_maxval, seed)
    lhs_zeros = jnp.zeros(lhs.shape, dtype=lhs.dtype)
    rhs = rand_unif((3, 3, contr_n, feature_n), rhs_maxval, seed + 1)
    rhs_zeros = jnp.zeros(rhs.shape, dtype=rhs.dtype)

    aqt_conv_fn = aqt_conv.make_conv_general_dilated_with_qt(dg_raw_conv)
    kwargs = {
        "window_strides": (1, 1),
        "padding": "SAME",
        "dimension_numbers": fl._conv_dimension_numbers(lhs.shape),
    }

    lhs_q, _ = dg_raw_conv.dg_quantizer.lhs.quant(
        lhs, calibration_axes=[0, 1, 2, 3])
    rhs_q, _ = dg_raw_conv.dg_quantizer.rhs.quant(
        rhs, calibration_axes=[0, 1, 2])

    out_no_quant, _ = aqt_conv_fn(
        lhs, rhs, lhs_qt=None, rhs_qt=None, **kwargs)

    out_lhs_quant, _ = aqt_conv_fn(
        lhs_zeros, rhs, lhs_qt=lhs_q, rhs_qt=None, **kwargs)
    out_rhs_quant, _ = aqt_conv_fn(
        lhs, rhs_zeros, lhs_qt=None, rhs_qt=rhs_q, **kwargs)
    out_both_quant, _ = aqt_conv_fn(
        lhs_zeros, rhs_zeros, lhs_qt=lhs_q, rhs_qt=rhs_q, **kwargs
    )
    assert (out_no_quant == out_lhs_quant).all()
    assert (out_no_quant == out_rhs_quant).all()
    assert (out_no_quant == out_both_quant).all()


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mnist example for intercept method API."""

from absl import app
from aqt.jax.v2 import config as aqt_config
from aqt.jax.v2 import tiled_dot_general
from aqt.jax.v2 import utils as aqt_utils
from aqt.jax.v2.examples import flax_e2e_model
from aqt.jax.v2.flax import aqt_flax
from aqt.jax.v2.flax.intercept import aqt_intercept_methods
from flax import linen as nn
from flax.metrics import tensorboard
import jax
import jax.numpy as jnp
import optax


class CNN(nn.Module):
  """A simple CNN model."""
  bn_use_stats: bool

  @nn.compact
  def __call__(self, x):
    use_running_avg = not self.bn_use_stats
    x = nn.Conv(features=32, kernel_size=(3, 3))(x)
    x = nn.BatchNorm(use_running_average=use_running_avg, dtype=x.dtype)(x)
    x = nn.relu(x)
    x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
    x = nn.Conv(features=64, kernel_size=(3, 3))(x)
    x = nn.BatchNorm(use_running_average=use_running_avg, dtype=x.dtype)(x)
    x = nn.relu(x)
    x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
    x = x.reshape((x.shape[0], -1))  # flatten
    x = nn.Dense(features=256)(x)
    x = nn.relu(x)
    x = nn.Dense(features=10)(x)

    # TODO(kimjaehong): Support einsum with the intercept method.
    return x


def create_train_state(rng):
  """Creates initial `flax_e2e_model.TrainState`."""
  cnn_train = CNN(bn_use_stats=True)
  model = cnn_train.init({'params': rng}, jnp.ones([1, 28, 28, 1]))
  learning_rate = 0.1
  momentum = 0.9
  tx = optax.sgd(learning_rate, momentum)
  cnn_eval = CNN(bn_use_stats=False)
  return flax_e2e_model.TrainState(
      cnn_train=cnn_train,
      cnn_eval=cnn_eval,
      model=model,
      tx=tx,
      opt_state=tx.init(model['params']),
  )


def train_and_evaluate(
    num_epochs: int, workdir: str, aqt_cfg: aqt_config.DotGeneral,
    tiling_cfg: tiled_dot_general.Cfg,
) -> flax_e2e_model.TrainState:
  """Execute model training and evaluation loop."""
  train_ds, test_ds = flax_e2e_model.get_datasets()
  rng = jax.random.key(0)

  summary_writer = tensorboard.SummaryWriter(workdir)

  rng, init_rng = jax.random.split(rng)
  state = create_train_state(init_rng)

  batch_size = 128
  with aqt_intercept_methods.intercept_methods(aqt_cfg, tiling_cfg=tiling_cfg):
    for epoch in range(1, num_epochs + 1):
      rng, input_rng = jax.random.split(rng)
      state, train_loss, train_accuracy = flax_e2e_model.train_epoch(
          state, train_ds, batch_size, input_rng
      )
      _, test_loss, test_accuracy, _ = flax_e2e_model.apply_model(
          state, test_ds['image'], test_ds['label'], train=False
      )

      print(
          'epoch:% 3d, train_loss: %.30f, train_accuracy: %.30f, test_loss:'
          ' %.30f, test_accuracy: %.30f'
          % (
              epoch,
              train_loss,
              train_accuracy * 100,
              test_loss,
              test_accuracy * 100,
          ),
          flush=True,
      )

      summary_writer.scalar('train_loss', train_loss, epoch)
      summary_writer.scalar('train_accuracy', train_accuracy, epoch)
      summary_writer.scalar('test_loss', test_loss, epoch)
      summary_writer.scalar('test_accuracy', test_accuracy, epoch)

    summary_writer.flush()
  return state


def serving_conversion(train_state, aqt_cfg, tiling_cfg):
  """Model conversion (quantized weights freezing)."""
  input_shape = (1, 28, 28, 1)
  cnn_freeze = CNN(
      bn_use_stats=False,
  )
  with aqt_intercept_methods.intercept_methods(
      aqt_cfg,
      rhs_quant_mode=aqt_utils.QuantMode.CONVERT,
      lhs_freeze_mode=aqt_flax.FreezerMode.NONE,
      rhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION_AND_VALUE,
      tiling_cfg=tiling_cfg):
    _, model_serving = cnn_freeze.apply(
        train_state.model,
        jnp.ones(input_shape),
        rngs={'params': jax.random.PRNGKey(0)},
        mutable=True,
    )
  cnn_serve = CNN(
      bn_use_stats=False,
  )

  serve_fn = aqt_intercept_methods.intercept_wrapper(
      cnn_serve.apply,
      aqt_cfg,
      rhs_quant_mode=aqt_utils.QuantMode.SERVE,
      lhs_freeze_mode=aqt_flax.FreezerMode.NONE,
      rhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION_AND_VALUE,
      tiling_cfg=tiling_cfg)
  return serve_fn, model_serving


@jax.jit
def serve(state,
          aqt_cfg: aqt_config.DotGeneral,
          tiling_cfg: tiled_dot_general.Cfg):
  """Take train state, freeze integer weights, and serve."""
  # get sample serving data
  _, test_ds = flax_e2e_model.get_datasets()
  sample_image, sample_label = test_ds['image'][:64], test_ds['label'][:64]
  # serving
  serve_fn, model_serving = serving_conversion(state, aqt_cfg, tiling_cfg)
  logits = serve_fn(
      model_serving, sample_image, rngs={'params': jax.random.PRNGKey(0)}
  )
  # compute serving loss
  one_hot = jax.nn.one_hot(sample_label, 10)
  loss = jnp.mean(optax.softmax_cross_entropy(logits=logits, labels=one_hot))
  return loss


def main(argv):
  del argv
  aqt_cfg = aqt_config.fully_quantized(fwd_bits=8, bwd_bits=8)
  tiling_cfg = tiled_dot_general.Cfg(
      lhs=tiled_dot_general.TensorTiling(
          contraction_axes=[
              tiled_dot_general.AxisTiling(
                  axis=1, tile_count=2, tile_size=None
              ),
          ],
          remaining_axes=[],
      ),
      rhs=tiled_dot_general.TensorTiling(
          contraction_axes=[
              tiled_dot_general.AxisTiling(
                  axis=0, tile_count=2, tile_size=None
              ),
          ],
          remaining_axes=[],
      ),
  )
  state = train_and_evaluate(
      num_epochs=2, workdir='/tmp/aqt_mnist_example',
      aqt_cfg=aqt_cfg, tiling_cfg=tiling_cfg
  )
  loss = serve(state, aqt_cfg, tiling_cfg)
  print('serve loss on sample ds: {}'.format(loss))


if __name__ == '__main__':
  app.run(main)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common utility functions used across different tests for both TF and JAX."""

import copy
from typing import List, Tuple, Union

from absl import flags
from aqt.common import aqt_common
from aqt.common import aqt_config
import numpy as np

flags.DEFINE_bool("use_tpu", False, "Set it to enable TPU-only tests.")


def test_stats_config():
  """Config base for all the experiments."""
  return aqt_config.StatsConfig(
      ema_update_count=1,  # single update overwrites the stats
      update_count_prior=0,  # easier equations
      share_stats_axes=[0, 1],  # one stat per whole tensor
      tpu_cross_replica_sum=False,  # no cross-tpu reduction
      filter_zeros=False,  # on default zeros are a number like any other
  )


def generate_unaligned_schedule_intervals():
  """Returns [start, stop) intervals for unaligned quantization schedules."""

  i64min = np.iinfo(np.int64).min
  i64max = np.iinfo(np.int64).max

  return [{
      "lhs_intervals": [(-1, 5)],
      "rhs_intervals": [(-1, 6)],
      "testcase_name": "ne_end_at_event"
  }, {
      "lhs_intervals": [(-1, 5)],
      "rhs_intervals": [(-2, 5)],
      "testcase_name": "ne_begin_at_event"
  }, {
      "lhs_intervals": [(None, None)],
      "rhs_intervals": [(None, i64max)],
      "testcase_name": "max_edge"
  }, {
      "lhs_intervals": [(None, None)],
      "rhs_intervals": [(i64min, None)],
      "testcase_name": "min_edge"
  }, {
      "lhs_intervals": [(0, 10)],
      "rhs_intervals": [(0, 5), (5, 10)],
      "testcase_name": "unequal_length"
  }]


def exact_int8_example(
    lhs_shape: Union[int, Tuple[int, ...]],  #
    rhs_shape: Union[int, Tuple[int, ...]],
    lhs_share_stats_axes: List[int],
    rhs_share_stats_axes: List[int],
    lhs_use_quantized_variable: bool,
    rhs_use_quantized_variable: bool
) -> Tuple[aqt_config.AqtScheduleConfig, np.ndarray,
           aqt_config.AqtScheduleConfig, np.ndarray]:
  """Returns a pair of matrices and config to multiply exactly.

  Exact int8 testing examples are used across different tests in AQTp with
  different specs required for creating examples, in order to validate the
  correctness of quantization ops. This method creates examples tensors and
  their configs as specified by the arguments depending on use cases.

  Args:
    lhs_shape: Shape of a tensor to be created for LHS.
    rhs_shape: Shape of a tensor to be created for RHS.
    lhs_share_stats_axes: Contraction axes for LHS.
    rhs_share_stats_axes: Contraction axes for RHS.
    lhs_use_quantized_variable:
      For LHS, if true and in training mode, saves intermediate quantizations to
      user-provided variables. During inference, quantized variables
      are read from but not written to with new input values.
    rhs_use_quantized_variable:
      For RHS, same as above.
  Returns:
    AqtScheduleConfigs for both LHS and RHS, and their corresponding tensors.
  """

  rng = np.random.default_rng(1234)
  bits = 8

  # A subset of the range of numbers which can be preserved exactly.
  symmetric_uniform_range = 2**(bits - 1) - 1
  lo, hi = -symmetric_uniform_range, symmetric_uniform_range

  lhs = rng.integers(lo, hi, size=lhs_shape, dtype=np.int64)
  rhs = rng.integers(lo, hi, size=rhs_shape, dtype=np.int64)

  iqc = aqt_config.IntQuantConfig(bits=8)
  clip_bound = aqt_common.get_clip_bound(iqc)
  assert symmetric_uniform_range <= clip_bound

  sc = aqt_config.StatsConfig(
      ema_update_count=10,
      share_stats_axes=lhs_share_stats_axes,
      tpu_cross_replica_sum=False)
  tc = aqt_config.AqtTensorConfig(
      freeze_scale_at_begin=True,
      quant_config=iqc,
      calibration_config=aqt_config.CalibrationConfig(
          const_bound_coeff=clip_bound))

  lhs_config = aqt_config.AqtScheduleConfig(sc, [tc],
                                            lhs_use_quantized_variable)
  rhs_config = copy.deepcopy(lhs_config)
  rhs_config.stats_config.share_stats_axes = rhs_share_stats_axes
  rhs_config.use_quantized_variable = rhs_use_quantized_variable

  lhs = np.array(lhs, np.float32)
  rhs = np.array(rhs, np.float32)

  return lhs_config, lhs, rhs_config, rhs



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""AQT APIs for pallas."""

import aqt.jax.v2.pallas.dot_general as _aqt_dot_general
import aqt.jax.v2.pallas.pallas_call as _aqt_pallas_call
import aqt.jax.v2.pallas.quantizer as _aqt_quantizer

pallas_call = _aqt_pallas_call.pallas_call
quant = _aqt_quantizer.quant
load_qtensor = _aqt_dot_general.load_qtensor
dot_general = _aqt_dot_general.dot_general
DequantMode = _aqt_dot_general.DequantMode



# UPDATE ME
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for AQT flax."""

import functools
from typing import Callable

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import aqt_dot_general
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import config
from aqt.jax.v2 import tiled_dot_general
from aqt.jax.v2 import utils
from aqt.jax.v2.flax import aqt_flax
import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np


class AqtFlaxTest(parameterized.TestCase):

  def test_aqt_einsum(self):
    class Model(nn.Module):
      aqt_cfg: None | Callable[[], config.DotGeneral]
      lhs_qt_external: bool = False
      rhs_qt_external: bool = False

      @nn.compact
      def __call__(self, lhs, rhs):
        lhs_in, rhs_in = lhs, rhs
        assert not (
            (self.lhs_qt_external or self.rhs_qt_external)
            and (self.aqt_cfg is None)
        ), 'aqt_cfg cannot be None when providing qtensor as inputs to einsum'
        aqt_cfg = self.aqt_cfg() if self.aqt_cfg is not None else None
        if self.lhs_qt_external or self.rhs_qt_external:
          # We are calling a config instantiation here again to get an
          # copy of the dg_quantized for this test setup that does not share
          # state with the one used in the einsum.
          dg_quantizer = self.aqt_cfg().fwd.dg_quantizer
          dg_quantizer.init_calibration()
          dimension_numbers = (((2, 3), (1, 2)), ((), ()))
          (lhs_q, _), (rhs_q, _) = dg_quantizer(
              lhs,
              rhs,
              dimension_numbers,
              aqt_dot_general.CalibrationMode.CONTRACTING_AXIS,
              aqt_dot_general.CalibrationMode.CONTRACTING_AXIS,
          )
          lhs_dtype = aqt_cfg.fwd.dg_quantizer.lhs.numerics.get_dtype()
          rhs_dtype = aqt_cfg.fwd.dg_quantizer.rhs.numerics.get_dtype()
          if self.lhs_qt_external:
            lhs_in = lhs_q.qvalue_astype(lhs_dtype)
          if self.rhs_qt_external:
            rhs_in = rhs_q.qvalue_astype(rhs_dtype)

        einsum = aqt_flax.AqtEinsum(
            cfg=aqt_cfg,
            lhs_freeze_mode=aqt_flax.FreezerMode.NONE,
            rhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION_AND_VALUE,
        )
        # xhs_qt can be inputs to AqtEinsum
        # xhs->xhs_qt can happen outside of AqtEinsum, e.g., k/v cache quant
        # input xhs_qt will force get_tensor() to always return xhs_qt
        out = einsum('ijkh,mkh->ijm', lhs_in, rhs_in)
        return out

    key = jax.random.PRNGKey(0)
    subkey1, subkey2 = jax.random.split(key, num=2)
    lhs = jax.random.uniform(subkey1, shape=(3, 4, 5, 6))
    rhs = jax.random.uniform(subkey2, shape=(2, 5, 6))

    def test(model_cls, aqt_cfg):
      model = model_cls(aqt_cfg=aqt_cfg)
      out = model.apply({}, lhs, rhs, rngs={'params': jax.random.PRNGKey(0)})
      # print(f'{out.shape=}')
      # print(f'{out=}')
      return out

    out_float = test(Model, lambda: None)
    out_int8 = test(Model, config.config_v4)
    out_int8_lqt = test(
        functools.partial(Model, lhs_qt_external=True),
        config.config_v4,
    )
    out_int8_rqt = test(
        functools.partial(Model, rhs_qt_external=True),
        config.config_v4,
    )
    out_int8_qt = test(
        functools.partial(Model, lhs_qt_external=True, rhs_qt_external=True),
        config.config_v4,
    )

    assert (out_int8 == out_int8_lqt).all(), 'lhs external qt failed'
    assert (out_int8 == out_int8_rqt).all(), 'rhs external qt failed'
    assert (out_int8 == out_int8_qt).all(), 'both external qt failed'
    mse = jnp.mean(jnp.square(out_int8_qt - out_float))
    assert mse > 0, 'Mean square error is 0. Einsum is not quantized.'

  def test_einsum_grad_leak(self):
    class CNN(nn.Module):
      aqt_cfg: config.DotGeneral

      @nn.compact
      def __call__(self, x):
        einsum = aqt_flax.AqtEinsum(
            self.aqt_cfg,
            lhs_freeze_mode=aqt_flax.FreezerMode.NONE,
            rhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION_AND_VALUE
        )
        x = einsum('bc,ab->ac', jnp.identity(10, dtype=x.dtype), x)
        return x

    model = CNN(aqt_cfg=config.fully_quantized())
    var = model.init({'params': jax.random.PRNGKey(0)}, jnp.ones(shape=(1, 10)))

    @jax.jit
    def apply_fn(inputs):
      return model.apply(
          var, inputs, rngs={'params': jax.random.PRNGKey(0)}, mutable=True
      )

    @jax.jit
    @jax.value_and_grad
    def train_step(inputs):
      return jnp.sum(apply_fn(inputs)[0])

    train_step(jnp.ones(shape=(1, 10)))

  @parameterized.parameters(True, False)
  def test_freezer(self, use_legacy_freezer: bool):
    class Model(nn.Module):
      aqt_cfg: None | config.DotGeneral
      lhs_quant_mode: aqt_flax.QuantMode
      rhs_quant_mode: aqt_flax.QuantMode
      use_legacy_freezer: bool

      @nn.compact
      def __call__(self, lhs):
        kernel = self.param(
            'kernel', nn.initializers.lecun_normal(), shape=(2, 5, 6)
        )
        einsum = aqt_flax.AqtEinsum(
            cfg=self.aqt_cfg,
            lhs_quant_mode=self.lhs_quant_mode,
            rhs_quant_mode=self.rhs_quant_mode,
            lhs_freeze_mode=aqt_flax.FreezerMode.NONE,
            rhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION_AND_VALUE,
            use_legacy_freezer=self.use_legacy_freezer,
        )
        out = einsum('ijkh,mkh->ijm', lhs, kernel)
        return out

    key = jax.random.PRNGKey(0)
    subkeys = jax.random.split(key, num=6)
    lhs = jax.random.uniform(subkeys[0], shape=(3, 4, 5, 6))

    aqt_cfg = config.config_v4()
    train_model = Model(
        aqt_cfg,
        lhs_quant_mode=aqt_flax.QuantMode.TRAIN,
        rhs_quant_mode=aqt_flax.QuantMode.TRAIN,
        use_legacy_freezer=use_legacy_freezer,
    )

    # Consider the initialized params as the already-trained params,
    # since we are testing Freezers which do nothing during training.
    train_model_params = train_model.init(subkeys[1], lhs)
    inference_result = train_model.apply(
        train_model_params, lhs, rngs={'params': subkeys[2]}
    )

    # Convert test.
    convert_model = Model(
        aqt_cfg,
        lhs_quant_mode=aqt_flax.QuantMode.TRAIN,
        rhs_quant_mode=aqt_flax.QuantMode.CONVERT,
        use_legacy_freezer=use_legacy_freezer,
    )
    _, converted_params = convert_model.apply(
        train_model_params, lhs, mutable=True, rngs={'params': subkeys[3]}
    )

    aqt_params = converted_params['aqt']
    w = converted_params['params']['kernel']
    w_qtensor = aqt_params['AqtEinsum_0']['AqtDotGeneral_0']['qrhs']
    if use_legacy_freezer:
      w_quant = w_qtensor['value']
    else:
      w_quant = w_qtensor['frozen'].qvalue

    self.assertEqual(w.shape, w_quant.shape)
    self.assertEqual(w_quant.dtype, jnp.int8)

    # Serve test.
    serve_model = Model(
        aqt_cfg,
        lhs_quant_mode=aqt_flax.QuantMode.TRAIN,
        rhs_quant_mode=aqt_flax.QuantMode.SERVE,
        use_legacy_freezer=use_legacy_freezer,
    )
    serve_model_init = serve_model.init(subkeys[4], lhs)

    # See if the QTensors are initialized properly to handle the general serving
    # pipeline of: initialize --> replace values with loaded checkpoint --> run
    init_aqt_params = serve_model_init['aqt']
    init_qtensor = init_aqt_params['AqtEinsum_0']['AqtDotGeneral_0']['qrhs']

    init_leaves, init_treedef = jax.tree.flatten(init_qtensor)
    converted_leaves, converted_treedef = jax.tree.flatten(w_qtensor)

    # 1. Same treestructure.
    self.assertEqual(init_treedef, converted_treedef)

    # 2. Leaves with the same shapes and dtypes. Values could be different.
    self.assertEqual(len(init_leaves), len(converted_leaves))
    for init_leaf, converted_leaf in zip(init_leaves, converted_leaves):
      self.assertEqual(init_leaf.shape, converted_leaf.shape)
      self.assertEqual(init_leaf.dtype, converted_leaf.dtype)

    # Serving test with converted params.
    quantized_result = serve_model.apply(
        converted_params, lhs, rngs={'params': subkeys[5]}
    )

    np.testing.assert_allclose(inference_result, quantized_result)

  def test_dot_general_tiling_fn(self):
    def _tiling_fn(lhs, rhs, dimension_numbers, tile_size=2):
      del lhs, rhs
      (lhs_ca, rhs_ca), _ = dimension_numbers
      ret = tiled_dot_general.Cfg(
          lhs=tiled_dot_general.TensorTiling(
              contraction_axes=[], remaining_axes=[]
          ),
          rhs=tiled_dot_general.TensorTiling(
              contraction_axes=[], remaining_axes=[]
          ),
      )
      for lhs_idx, rhs_idx in zip(lhs_ca, rhs_ca):
        ret.lhs.contraction_axes.append(
            tiled_dot_general.AxisTiling(
                axis=lhs_idx, tile_size=tile_size, tile_count=None
            )
        )
        ret.rhs.contraction_axes.append(
            tiled_dot_general.AxisTiling(
                axis=rhs_idx, tile_size=tile_size, tile_count=None
            )
        )

      return ret

    class Model(nn.Module):
      aqt_cfg: None | config.DotGeneral
      lhs_quant_mode: aqt_flax.QuantMode
      rhs_quant_mode: aqt_flax.QuantMode

      @nn.compact
      def __call__(self, lhs):
        kernel = self.param(
            'kernel', nn.initializers.lecun_normal(), shape=(2, 6, 10)
        )
        dot_general = aqt_flax.AqtDotGeneral(
            cfg=self.aqt_cfg,
            lhs_quant_mode=self.lhs_quant_mode,
            rhs_quant_mode=self.rhs_quant_mode,
            lhs_freeze_mode=aqt_flax.FreezerMode.NONE,
            rhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION_AND_VALUE,
            tiling_fn=_tiling_fn,
            use_legacy_freezer=False
        )
        out = dot_general(lhs, kernel, (((1, 2), (1, 2)), ((), ())), None)
        return out

    key = jax.random.PRNGKey(0)
    subkeys = jax.random.split(key, num=6)
    lhs = jax.random.uniform(subkeys[0], shape=(7, 6, 10))

    aqt_cfg = config.config_v4()
    train_model = Model(
        aqt_cfg,
        lhs_quant_mode=aqt_flax.QuantMode.TRAIN,
        rhs_quant_mode=aqt_flax.QuantMode.TRAIN
    )

    # Consider the initialized params as the already-trained params,
    # since we are testing Freezers which do nothing during training.
    train_model_params = train_model.init(subkeys[1], lhs)

    # Convert test.
    convert_model = Model(
        aqt_cfg,
        lhs_quant_mode=aqt_flax.QuantMode.TRAIN,
        rhs_quant_mode=aqt_flax.QuantMode.CONVERT,
    )
    _, converted_params = convert_model.apply(
        train_model_params, lhs, mutable=True, rngs={'params': subkeys[3]}
    )

    dtype = jnp.dtype
    expected_pytree = {
        'AqtDotGeneral_0': {
            'qrhs': {
                'frozen': aqt_tensor.QTensor(
                    qvalue=(dtype('int8'), (2, 3, 2, 5, 2)),
                    scale=[(dtype('float32'), (2, 3, 1, 5, 1))],
                    scale_t=None,
                    bias=[],
                    dequant_dtype=dtype('float32'),
                )
            }
        }
    }

    converted_pytree = jax.tree.map(
        lambda x: (x.dtype, x.shape), converted_params['aqt']
    )
    utils.test_pprint_eq(expected_pytree, converted_pytree)

  def test_einsum_tiling_fn(self):
    def _tiling_fn(eqn, lhs, rhs, tile_size=2):
      del lhs, rhs
      return tiled_dot_general.Cfg.from_einsum(eqn, {'d': tile_size})

    class Model(nn.Module):
      aqt_cfg: None | config.DotGeneral
      lhs_quant_mode: aqt_flax.QuantMode
      rhs_quant_mode: aqt_flax.QuantMode

      @nn.compact
      def __call__(self, lhs):
        kernel = self.param(
            'kernel', nn.initializers.lecun_normal(), shape=(3, 6, 10)
        )
        einsum = aqt_flax.AqtEinsum(
            cfg=self.aqt_cfg,
            lhs_quant_mode=self.lhs_quant_mode,
            rhs_quant_mode=self.rhs_quant_mode,
            lhs_freeze_mode=aqt_flax.FreezerMode.NONE,
            rhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION_AND_VALUE,
            tiling_fn=_tiling_fn,
            use_legacy_freezer=False
        )
        out = einsum('abd,cbd->ac', lhs, kernel)
        return out

    key = jax.random.PRNGKey(0)
    subkeys = jax.random.split(key, num=6)
    lhs = jax.random.uniform(subkeys[0], shape=(7, 6, 10))

    aqt_cfg = config.config_v4()
    train_model = Model(
        aqt_cfg,
        lhs_quant_mode=aqt_flax.QuantMode.TRAIN,
        rhs_quant_mode=aqt_flax.QuantMode.TRAIN
    )

    # Consider the initialized params as the already-trained params,
    # since we are testing Freezers which do nothing during training.
    train_model_params = train_model.init(subkeys[1], lhs)

    # Convert test.
    convert_model = Model(
        aqt_cfg,
        lhs_quant_mode=aqt_flax.QuantMode.TRAIN,
        rhs_quant_mode=aqt_flax.QuantMode.CONVERT,
    )
    _, converted_params = convert_model.apply(
        train_model_params, lhs, mutable=True, rngs={'params': subkeys[3]}
    )

    dtype = jnp.dtype
    expected_pytree = {
        'AqtEinsum_0': {
            'AqtDotGeneral_0': {
                'qrhs': {
                    'frozen': aqt_tensor.QTensor(
                        qvalue=(dtype('int8'), (3, 6, 5, 2)),
                        scale=[(dtype('float32'), (3, 1, 5, 1))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype('float32'),
                    )
                }
            }
        }
    }

    converted_pytree = jax.tree.map(
        lambda x: (x.dtype, x.shape), converted_params['aqt']
    )
    utils.test_pprint_eq(expected_pytree, converted_pytree)

if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for calling pallas functions from JAX."""

import collections
from typing import Any, Callable

from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2.pallas import pallas_tensor
import jax
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu


BlockSpec = pl.BlockSpec
no_block_spec = pl.no_block_spec
tree_util = jax.tree_util

QTensor = aqt_tensor.QTensor
TransposedTensor = pallas_tensor.TransposedTensor

ArgAndBlockSpec = collections.namedtuple('ArgAndBlockSpec', 'arg block_spec')


def _make_qtensor_blockspec(arg, block_spec) -> QTensor | Any:
  if isinstance(arg, QTensor) and isinstance(block_spec, BlockSpec):
    return pallas_tensor.make_qtensor_blockspec(arg, block_spec)
  else:
    return block_spec


def _transpose_tensor_for_memory_saving(
    arg: Any, block_spec: BlockSpec
) -> ArgAndBlockSpec:
  """Transposes tensor for memory optimization."""
  arg, block_spec = pallas_tensor.transpose_tensor_for_memory_saving(
      arg, block_spec
  )
  return ArgAndBlockSpec(arg, block_spec)


def _is_qtensor(x):
  return isinstance(x, QTensor)


def _is_transposed_tensor(x):
  return isinstance(x, TransposedTensor)


def _is_arg_and_block_spec(x):
  return isinstance(x, ArgAndBlockSpec)


def pallas_call(
    f: Callable[..., None],
    *pl_call_args,
    grid_spec=None,
    in_specs=no_block_spec,
    **pl_call_kwrags,
):
  """pl.pallas_call wrapper that can pass QTensor as input."""

  # If grid spec is given, use in_specs from grid spec.
  if grid_spec is not None:
    if jax.__version_info__ >= (0, 4, 31):
      in_specs = grid_spec.in_specs
    else:
      in_specs = tree_util.tree_unflatten(
          grid_spec.in_specs_tree, grid_spec.in_specs
      )

  @jax.jit
  def wrapped(*args):

    prefetch_args = ()
    if isinstance(grid_spec, pltpu.PrefetchScalarGridSpec):
      prefetch_args, args = (
          args[: grid_spec.num_scalar_prefetch],
          args[grid_spec.num_scalar_prefetch :],
      )

    # Flatten args and its inspecs.
    flat_args, args_treedef = tree_util.tree_flatten(
        args, is_leaf=_is_qtensor
    )
    flat_inspecs, inspecs_treedef = tree_util.tree_flatten(in_specs)

    # Build block spec for each QTensor in arguments.
    flat_inspecs = tree_util.tree_map(
        _make_qtensor_blockspec,
        flat_args,
        flat_inspecs,
        is_leaf=_is_qtensor,
    )

    # Transpose tensor for memory optimization.
    flat_args_and_inspecs = tree_util.tree_map(
        _transpose_tensor_for_memory_saving,
        flat_args,
        flat_inspecs,
    )

    flat_args = tree_util.tree_map(
        lambda x: x.arg, flat_args_and_inspecs, is_leaf=_is_arg_and_block_spec
    )
    flat_inspecs = tree_util.tree_map(
        lambda x: x.block_spec,
        flat_args_and_inspecs,
        is_leaf=_is_arg_and_block_spec,
    )

    # Unflatten args and its inspecs.
    args = tree_util.tree_unflatten(
        args_treedef, flat_args
    )
    kernel_inspecs = tree_util.tree_unflatten(
        inspecs_treedef, flat_inspecs
    )

    def kernel(*args):
      # [Inside kernel] untranspose tensor to restore to the original shape.
      args = jax.tree_util.tree_map(
          lambda arg: arg.untransposed
          if isinstance(arg, TransposedTensor)
          else arg,
          args,
          is_leaf=_is_transposed_tensor,
      )
      return f(*args)

    if grid_spec is not None:
      if jax.__version_info__ >= (0, 4, 31):
        grid_spec.in_specs = kernel_inspecs
      else:
        kernel_inspecs, inspecs_treedef = tree_util.tree_flatten(kernel_inspecs)
        grid_spec.in_specs = tuple(kernel_inspecs)
        grid_spec.in_specs_tree = inspecs_treedef

    func = pl.pallas_call(
        kernel,
        *pl_call_args,
        grid_spec=grid_spec,
        in_specs=kernel_inspecs if grid_spec is None else no_block_spec,
        **pl_call_kwrags,
    )
    return func(*prefetch_args, *args)

  return wrapped



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""AQT Quantizr API."""

# TODO(wppark): Remove this file. This is a temporary module before the
# official release of AQT quant / dequant API.

from typing import Sequence

from aqt.jax.v2 import aqt_quantizer
from aqt.jax.v2 import aqt_tensor

import jax
import jax.numpy as jnp

QTensor = aqt_tensor.QTensor


def quant(
    x: jax.Array,
    n_bits: int,
    calibration_axes: Sequence[int],
) -> QTensor:
  """Apply channel-wise quantization to x.

  x is quantized channel-wisely on the calibration axes.

  Args:
    x: input tensor
    n_bits: the precision for quantization.
    calibration_axes: the calibration axes.
  Returns:
    A quantized QTensor
  """
  # jax.lax.stop_gradient is not supported in pallas, thus disable
  # scale_stop_grad in the quantizer.
  # VPU ops only support float32. jax implicitly converts tensor into float32.
  # However, pallas requires explicit casting. Therefore, we need to enforce
  # the scale factor and dequant dtype to be float32.
  quantizer = aqt_quantizer.quantizer_make(
      n_bits, scale_stop_grad=False, scale_dtype=jnp.float32
  )

  qx, _ = quantizer.quant(x, calibration_axes=calibration_axes)
  qx.dequant_dtype = jnp.float32
  return qx
